from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ConsentStatus,
    ConsentType,
    ExtractedConsentIdentifier,
    MergedConsentIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    YearMonthDayTime,
)


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Consent"], Field(frozen=True)]] = "Consent"


class _OptionalValues(_Stem):
    hasConsentType: Annotated[
        ConsentType | None,
        Field(
            description="The type of consent.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = None


class _RequiredValues(_Stem):
    hasConsentStatus: Annotated[
        ConsentStatus,
        Field(
            description="Specifies the state or status of consent.",
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasConsentStatus"]},
        ),
    ]
    hasDataSubject: Annotated[
        MergedPersonIdentifier,
        Field(
            description="Indicates association with Data Subject.",
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasDataSubject"]},
        ),
    ]
    isIndicatedAtTime: Annotated[
        YearMonthDayTime,
        Field(
            description=(
                "Specifies the temporal information for when the entity has "
                "indicated the specific context."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#isIndicatedAtTime"]},
        ),
    ]


class _SparseValues(_Stem):
    hasConsentStatus: Annotated[
        ConsentStatus | None,
        Field(
            description="Specifies the state or status of consent.",
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasConsentStatus"]},
        ),
    ] = None
    hasDataSubject: Annotated[
        MergedPersonIdentifier | None,
        Field(
            description="Indicates association with Data Subject.",
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasDataSubject"]},
        ),
    ] = None
    hasConsentType: Annotated[
        ConsentType | None,
        Field(
            description="The type of consent.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = None
    isIndicatedAtTime: Annotated[
        YearMonthDayTime | None,
        Field(
            description=(
                "Specifies the temporal information for when the entity has "
                "indicated the specific context."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#isIndicatedAtTime"]},
        ),
    ] = None


class _VariadicValues(_Stem):
    hasConsentType: Annotated[
        list[ConsentType],
        Field(
            description="The type of consent.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []
    hasConsentStatus: Annotated[
        list[ConsentStatus],
        Field(
            description=(
                "In DPV there is no property for the type of consent. In DPV, the "
                "types are subclasses of dpv:Consent. In order to align with our "
                "modelling approach, we model types of consent as a concept "
                "(skos:Concept). Since dpv:Consent is also defined as a "
                "skos:Concept, this is no conflict, just another way of "
                "implementing it."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasConsentStatus"]},
        ),
    ] = []
    hasDataSubject: Annotated[
        list[MergedPersonIdentifier],
        Field(
            description="Indicates association with Data Subject.",
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#hasDataSubject"]},
        ),
    ] = []
    isIndicatedAtTime: Annotated[
        list[YearMonthDayTime],
        Field(
            description=(
                "Specifies the temporal information for when the entity has "
                "indicated the specific context."
            ),
            json_schema_extra={"sameAs": ["https://w3id.org/dpv#isIndicatedAtTime"]},
        ),
    ] = []


class BaseConsent(
    _OptionalValues,
    _RequiredValues,
    json_schema_extra={
        "description": "Consent of the Data Subject for specified process or activity.",
        "sameAs": ["https://w3id.org/dpv#Consent"],
    },
):
    """All fields for a valid consent except for provenance."""


class ExtractedConsent(BaseConsent, ExtractedData):
    """An automatically extracted metadata set describing a consent."""

    entityType: Annotated[
        Literal["ExtractedConsent"], Field(alias="$type", frozen=True)
    ] = "ExtractedConsent"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedConsentIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedConsentIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedConsentIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedConsentIdentifier)


class MergedConsent(BaseConsent, MergedItem):
    """The result of merging all extracted items and rules for a consent."""

    entityType: Annotated[
        Literal["MergedConsent"], Field(alias="$type", frozen=True)
    ] = "MergedConsent"
    identifier: Annotated[
        MergedConsentIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedConsentIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewConsent(_OptionalValues, _SparseValues, PreviewItem):
    """Preview for merging all extracted items and rules for a consent."""

    entityType: Annotated[
        Literal["PreviewConsent"], Field(alias="$type", frozen=True)
    ] = "PreviewConsent"
    identifier: Annotated[
        MergedConsentIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedConsentIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveConsent(_OptionalValues, _SparseValues, AdditiveRule):
    """Rule to add values to merged consent items."""

    entityType: Annotated[
        Literal["AdditiveConsent"], Field(alias="$type", frozen=True)
    ] = "AdditiveConsent"
    supersededBy: Annotated[
        MergedConsentIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveConsent(_VariadicValues, SubtractiveRule):
    """Rule to subtract values from merged consent items."""

    entityType: Annotated[
        Literal["SubtractiveConsent"], Field(alias="$type", frozen=True)
    ] = "SubtractiveConsent"


class PreventiveConsent(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged consent items."""

    entityType: Annotated[
        Literal["PreventiveConsent"], Field(alias="$type", frozen=True)
    ] = "PreventiveConsent"
    hasConsentType: list[MergedPrimarySourceIdentifier] = []
    hasConsentStatus: list[MergedPrimarySourceIdentifier] = []
    hasDataSubject: list[MergedPrimarySourceIdentifier] = []
    isIndicatedAtTime: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for a consent item."""

    additive: AdditiveConsent = AdditiveConsent()
    subtractive: SubtractiveConsent = SubtractiveConsent()
    preventive: PreventiveConsent = PreventiveConsent()


class ConsentRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a consent item."""

    entityType: Annotated[
        Literal["ConsentRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "ConsentRuleSetRequest"


class ConsentRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a consent item."""

    entityType: Annotated[
        Literal["ConsentRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "ConsentRuleSetResponse"
    stableTargetId: MergedConsentIdentifier


class ConsentMapping(_Stem, BaseMapping):
    """Mapping for describing a consent transformation."""

    entityType: Annotated[
        Literal["ConsentMapping"], Field(alias="$type", frozen=True)
    ] = "ConsentMapping"
    hasConsentStatus: Annotated[list[MappingField[ConsentStatus]], Field(min_length=1)]
    hasDataSubject: Annotated[
        list[MappingField[MergedPersonIdentifier]], Field(min_length=1)
    ]
    isIndicatedAtTime: Annotated[
        list[MappingField[YearMonthDayTime]], Field(min_length=1)
    ]
    hasConsentType: list[MappingField[ConsentType | None]] = []


class ConsentFilter(_Stem, BaseFilter):
    """Class for defining filter rules for consent items."""

    entityType: Annotated[
        Literal["ConsentFilter"], Field(alias="$type", frozen=True)
    ] = "ConsentFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from typing import cast

from mex.common.models import (
    ADDITIVE_MODEL_CLASSES_BY_NAME,
    EXTRACTED_MODEL_CLASSES_BY_NAME,
    MERGED_MODEL_CLASSES_BY_NAME,
    PREVENTIVE_MODEL_CLASSES_BY_NAME,
    PREVIEW_MODEL_CLASSES_BY_NAME,
    SUBTRACTIVE_MODEL_CLASSES_BY_NAME,
)
from mex.common.types import (
    MERGED_IDENTIFIER_CLASSES,
    NESTED_MODEL_CLASSES_BY_NAME,
    TEMPORAL_ENTITIES,
    VOCABULARY_ENUMS,
    AnyTemporalEntity,
    AnyVocabularyEnum,
    Link,
    LiteralStringType,
    TemporalEntityPrecision,
    Text,
)
from mex.common.utils import (
    contains_only_types,
    get_all_fields,
    get_inner_types,
    group_fields_by_class_name,
)

# all models classes
ALL_MODEL_CLASSES_BY_NAME = {
    **ADDITIVE_MODEL_CLASSES_BY_NAME,
    **EXTRACTED_MODEL_CLASSES_BY_NAME,
    **MERGED_MODEL_CLASSES_BY_NAME,
    **NESTED_MODEL_CLASSES_BY_NAME,
    **PREVIEW_MODEL_CLASSES_BY_NAME,
    **PREVENTIVE_MODEL_CLASSES_BY_NAME,
    **SUBTRACTIVE_MODEL_CLASSES_BY_NAME,
}

# all field types grouped by field and class names
ALL_TYPES_BY_FIELDS_BY_CLASS_NAMES = {
    class_name: {
        field_name: list(get_inner_types(field_info.annotation, include_none=False))
        for field_name, field_info in get_all_fields(model_class).items()
    }
    for class_name, model_class in ALL_MODEL_CLASSES_BY_NAME.items()
}

# fields that are immutable and can only be set once
FROZEN_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: field_info.frozen is True,
)

# static fields that are set once on class-level to a literal type
LITERAL_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: isinstance(field_info.annotation, LiteralStringType),
)

# fields typed as merged identifiers containing references to merged items
REFERENCE_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, *MERGED_IDENTIFIER_CLASSES),
)

# nested fields that contain `Text` objects
TEXT_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, Text),
)

# nested fields that contain `Link` objects
LINK_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, Link),
)

# fields annotated as `int` type
INTEGER_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, int),
)

# fields annotated as `str` type
STRING_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, str),
)

# fields annotated as any temporal type
TEMPORAL_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, *TEMPORAL_ENTITIES),
)

# allowed temporal precisions grouped by field and class names
TEMPORAL_PRECISIONS_BY_FIELD_BY_CLASS_NAMES = {
    class_name: {
        field_name: sorted(
            {
                precision
                for temporal_type in ALL_TYPES_BY_FIELDS_BY_CLASS_NAMES[class_name][
                    field_name
                ]
                for precision in cast(
                    "type[AnyTemporalEntity]", temporal_type
                ).ALLOWED_PRECISION_LEVELS
            },
            key=lambda precision: list(TemporalEntityPrecision).index(precision),
        )
        for field_name in field_names
    }
    for class_name, field_names in TEMPORAL_FIELDS_BY_CLASS_NAME.items()
}

# fields annotated as any vocabulary enum
VOCABULARY_FIELDS_BY_CLASS_NAME = group_fields_by_class_name(
    ALL_MODEL_CLASSES_BY_NAME,
    lambda field_info: contains_only_types(field_info, *VOCABULARY_ENUMS),
)

# vocabulary enum items grouped by field and class names
VOCABULARIES_BY_FIELDS_BY_CLASS_NAMES = {
    class_name: {
        field_name: [
            item
            for vocabulary in ALL_TYPES_BY_FIELDS_BY_CLASS_NAMES[class_name][field_name]
            for item in cast("type[AnyVocabularyEnum]", vocabulary)
        ]
        for field_name in field_names
    }
    for class_name, field_names in VOCABULARY_FIELDS_BY_CLASS_NAME.items()
}

# fields with changeable values that are not nested objects or merged item references
MUTABLE_FIELDS_BY_CLASS_NAME = {
    name: sorted(
        {
            field_name
            for field_name in get_all_fields(cls)
            if field_name
            not in (
                *FROZEN_FIELDS_BY_CLASS_NAME[name],
                *REFERENCE_FIELDS_BY_CLASS_NAME[name],
                *TEXT_FIELDS_BY_CLASS_NAME[name],
                *LINK_FIELDS_BY_CLASS_NAME[name],
            )
        }
    )
    for name, cls in ALL_MODEL_CLASSES_BY_NAME.items()
}

# fields with mergeable values that are neither literal nor frozen
MERGEABLE_FIELDS_BY_CLASS_NAME = {
    name: sorted(
        {
            field_name
            for field_name in get_all_fields(cls)
            if field_name
            not in (
                *FROZEN_FIELDS_BY_CLASS_NAME[name],
                *LITERAL_FIELDS_BY_CLASS_NAME[name],
            )
        }
    )
    for name, cls in ALL_MODEL_CLASSES_BY_NAME.items()
}

# fields with values that should be set once but are neither literal nor references
FINAL_FIELDS_BY_CLASS_NAME = {
    name: sorted(
        {
            field_name
            for field_name in get_all_fields(cls)
            if field_name in FROZEN_FIELDS_BY_CLASS_NAME[name]
            and field_name
            not in (
                *LITERAL_FIELDS_BY_CLASS_NAME[name],
                *REFERENCE_FIELDS_BY_CLASS_NAME[name],
            )
        }
    )
    for name, cls in ALL_MODEL_CLASSES_BY_NAME.items()
}

# list of fields by class name that are not allowed to be null or an empty list
REQUIRED_FIELDS_BY_CLASS_NAME = {
    name: sorted(
        {
            field_name
            for field_name, field_info in cls.model_fields.items()
            if field_info.is_required()
        }
    )
    for name, cls in ALL_MODEL_CLASSES_BY_NAME.items()
}



from pydantic import Field

from mex.common.models import BaseModel


class OrcidIdentifier(BaseModel):
    """Model class for OrcidID."""

    path: str
    uri: str


class OrcidEmail(BaseModel):
    """Model class for Orcid email."""

    email: list[str]


class OrcidEmails(BaseModel):
    """Model class for Orcid emails."""

    email: list[OrcidEmail]


class OrcidFamilyName(BaseModel):
    """Model class for orcid family names."""

    value: str


class OrcidGivenNames(BaseModel):
    """Model class for Orcid given names."""

    value: str


class OrcidName(BaseModel):
    """Model class for Orcid name."""

    family_name: OrcidFamilyName | None = Field(alias="family-name")
    given_names: OrcidGivenNames | None = Field(alias="given-names")
    visibility: str


class OrcidPerson(BaseModel):
    """Model class for Orcid person."""

    emails: OrcidEmails
    name: OrcidName


class OrcidRecord(BaseModel):
    """Model class for Orcid record."""

    orcid_identifier: OrcidIdentifier = Field(alias="orcid-identifier")
    person: OrcidPerson


class OrcidSearchItem(BaseModel):
    """Model class for a single search result item."""

    orcid_identifier: OrcidIdentifier = Field(alias="orcid-identifier")


class OrcidSearchResponse(BaseModel):
    """Model class for a search response."""

    num_found: int = Field(alias="num-found")
    result: list[OrcidSearchItem]



from typing import TypeVar

from networkx import DiGraph, NetworkXUnfeasible
from networkx import topological_sort as network_sort

from mex.common.exceptions import MExError
from mex.common.utils import ensure_list

ItemT = TypeVar("ItemT")


def topological_sort(
    items: list[ItemT],
    primary_key: str,
    *,
    parent_key: str | None = None,
    child_key: str | None = None,
) -> None:
    """Sort the given list of items in-place according to their topology.

    Items can refer to each other using key fields. A parent item can reference a child
    item by storing the child's `primary_key` in the parent's `child_key` field.
    Similarly, a child can reference its parent using the `parent_key` field.

    This can be useful for submitting items to the backend in the correct order.
    """
    graph: DiGraph[str] = DiGraph()
    for item in items:
        current_node = getattr(item, primary_key)
        graph.add_node(current_node)
        if parent_key:
            for parent_node in ensure_list(getattr(item, parent_key)):
                graph.add_edge(parent_node, current_node)
        if child_key:
            for child_node in ensure_list(getattr(item, child_key)):
                graph.add_edge(current_node, child_node)

    try:
        sorted_keys = list(network_sort(graph))
    except NetworkXUnfeasible as error:
        msg = "Found graph cycles while sorting."
        raise MExError(msg) from error

    items.sort(
        key=lambda item: (
            sorted_keys.index(getattr(item, primary_key)),
            getattr(item, primary_key),
        )
    )



from collections.abc import Iterable

from mex.common.types import AnyPrimitiveType, MergedPrimarySourceIdentifier

SourceAndValueList = list[tuple[MergedPrimarySourceIdentifier, AnyPrimitiveType]]
SourceAndValueIter = Iterable[tuple[MergedPrimarySourceIdentifier, AnyPrimitiveType]]
ValueList = list[AnyPrimitiveType]
SourceList = list[MergedPrimarySourceIdentifier]



from typing import Annotated, ClassVar, Literal

from pydantic import AfterValidator, Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    APIType,
    ExtractedAccessPlatformIdentifier,
    Identifier,
    Link,
    MergedAccessPlatformIdentifier,
    MergedContactPointIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    TechnicalAccessibility,
    Text,
)

AnyContactIdentifier = Annotated[
    MergedOrganizationalUnitIdentifier
    | MergedPersonIdentifier
    | MergedContactPointIdentifier,
    AfterValidator(Identifier),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["AccessPlatform"], Field(frozen=True)]] = (
        "AccessPlatform"
    )


class _OptionalLists(_Stem):
    alternativeTitle: Annotated[
        list[Text],
        Field(
            description="An alternative name for the access platform.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]},
        ),
    ] = []
    contact: Annotated[
        list[AnyContactIdentifier],
        Field(
            description="An agent that serves as a contact for the access platform.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#contactPoint"]},
        ),
    ] = []
    description: Annotated[
        list[Text],
        Field(
            description="A short description of the access platform.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/description"]},
        ),
    ] = []
    landingPage: Annotated[
        list[Link],
        Field(
            description=(
                "A Web page that can be navigated to in a Web browser to gain "
                "access to the catalog, a dataset, its distributions and/or "
                "additional information."
            ),
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#landingPage"]},
        ),
    ] = []
    title: Annotated[
        list[Text],
        Field(
            description="The name of the access platform.",
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ] = []
    unitInCharge: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            description=(
                "This property refers to agents who assume responsibility and "
                "accountability for the resource and its appropriate maintenance."
            ),
            json_schema_extra={"sameAs": ["http://dcat-ap.de/def/dcatde/maintainer"]},
        ),
    ] = []


class _OptionalValues(_Stem):
    endpointDescription: Annotated[
        Link | None,
        Field(
            description=(
                "A description of the services available via the end-points, "
                "including their operations, parameters etc."
            ),
            json_schema_extra={
                "sameAs": ["http://www.w3.org/ns/dcat#endpointDescription"]
            },
        ),
    ] = None
    endpointType: Annotated[
        APIType | None,
        Field(
            description="The type of endpoint, e.g. REST.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = None
    endpointURL: Annotated[
        Link | None,
        Field(
            description=(
                "The root location or primary endpoint of the service "
                "(a Web-resolvable IRI)"
            ),
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#endpointURL"]},
        ),
    ] = None


class _RequiredValues(_Stem):
    technicalAccessibility: Annotated[
        TechnicalAccessibility,
        Field(
            description=(
                "Indicates form if the platform can be accessed only within RKI "
                "network (internally) or if the platform is accessible publicly "
                "(externally)."
            ),
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ]


class _SparseValues(_Stem):
    technicalAccessibility: Annotated[
        TechnicalAccessibility | None,
        Field(
            description=(
                "Indicates form if the platform can be accessed only within RKI "
                "network (internally) or if the platform is accessible publicly "
                "(externally)."
            ),
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = None


class _VariadicValues(_Stem):
    endpointDescription: Annotated[
        list[Link],
        Field(
            description=(
                "A description of the services available via the end-points, "
                "including their operations, parameters etc."
            ),
        ),
    ] = []
    endpointType: Annotated[
        list[APIType],
        Field(
            description="The type of endpoint, e.g. REST.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []
    endpointURL: Annotated[
        list[Link],
        Field(
            description=(
                "The root location or primary endpoint of the service "
                "(a Web-resolvable IRI)"
            ),
        ),
    ] = []
    technicalAccessibility: Annotated[
        list[TechnicalAccessibility],
        Field(
            description=(
                "Indicates form if the platform can be accessed only within RKI "
                "network (internally) or if the platform is accessible publicly "
                "(externally)."
            ),
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]},
        ),
    ] = []


class BaseAccessPlatform(
    _OptionalLists,
    _OptionalValues,
    _RequiredValues,
    json_schema_extra={
        "description": (
            "A technical system or service that provides access to distributions or "
            "resources."
        ),
        "sameAs": ["http://www.w3.org/ns/dcat#DataService"],
    },
):
    """All fields for a valid access platform except for provenance."""


class ExtractedAccessPlatform(BaseAccessPlatform, ExtractedData):
    """An automatically extracted metadata item describing an access platform."""

    entityType: Annotated[
        Literal["ExtractedAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "ExtractedAccessPlatform"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedAccessPlatformIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedAccessPlatformIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedAccessPlatformIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedAccessPlatformIdentifier)


class MergedAccessPlatform(BaseAccessPlatform, MergedItem):
    """The result of merging all extracted items and rules for an access platform."""

    entityType: Annotated[
        Literal["MergedAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "MergedAccessPlatform"
    identifier: Annotated[
        MergedAccessPlatformIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedAccessPlatformIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewAccessPlatform(
    _OptionalLists, _OptionalValues, _SparseValues, PreviewItem
):
    """Preview for merging all extracted items and rules for an access platform."""

    entityType: Annotated[
        Literal["PreviewAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "PreviewAccessPlatform"
    identifier: Annotated[
        MergedAccessPlatformIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedAccessPlatformIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveAccessPlatform(
    _OptionalLists, _OptionalValues, _SparseValues, AdditiveRule
):
    """Rule to add values to merged access platform items."""

    entityType: Annotated[
        Literal["AdditiveAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "AdditiveAccessPlatform"
    supersededBy: Annotated[
        MergedAccessPlatformIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveAccessPlatform(_OptionalLists, _VariadicValues, SubtractiveRule):
    """Rule to subtract values from merged access platform items."""

    entityType: Annotated[
        Literal["SubtractiveAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "SubtractiveAccessPlatform"


class PreventiveAccessPlatform(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged access platform items."""

    entityType: Annotated[
        Literal["PreventiveAccessPlatform"], Field(alias="$type", frozen=True)
    ] = "PreventiveAccessPlatform"
    alternativeTitle: list[MergedPrimarySourceIdentifier] = []
    contact: list[MergedPrimarySourceIdentifier] = []
    description: list[MergedPrimarySourceIdentifier] = []
    endpointDescription: list[MergedPrimarySourceIdentifier] = []
    endpointType: list[MergedPrimarySourceIdentifier] = []
    endpointURL: list[MergedPrimarySourceIdentifier] = []
    landingPage: list[MergedPrimarySourceIdentifier] = []
    technicalAccessibility: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []
    unitInCharge: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for an access platform item."""

    additive: AdditiveAccessPlatform = AdditiveAccessPlatform()
    subtractive: SubtractiveAccessPlatform = SubtractiveAccessPlatform()
    preventive: PreventiveAccessPlatform = PreventiveAccessPlatform()


class AccessPlatformRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update an access platform item."""

    entityType: Annotated[
        Literal["AccessPlatformRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "AccessPlatformRuleSetRequest"


class AccessPlatformRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve an access platform item."""

    entityType: Annotated[
        Literal["AccessPlatformRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "AccessPlatformRuleSetResponse"
    stableTargetId: MergedAccessPlatformIdentifier


class AccessPlatformMapping(_Stem, BaseMapping):
    """Mapping for describing an access platform transformation."""

    entityType: Annotated[
        Literal["AccessPlatformMapping"], Field(alias="$type", frozen=True)
    ] = "AccessPlatformMapping"
    technicalAccessibility: Annotated[
        list[MappingField[TechnicalAccessibility]], Field(min_length=1)
    ]
    endpointDescription: list[MappingField[Link | None]] = []
    endpointType: list[MappingField[APIType | None]] = []
    endpointURL: list[MappingField[Link | None]] = []
    alternativeTitle: list[MappingField[list[Text]]] = []
    contact: list[MappingField[list[AnyContactIdentifier]]] = []
    description: list[MappingField[list[Text]]] = []
    landingPage: list[MappingField[list[Link]]] = []
    title: list[MappingField[list[Text]]] = []
    unitInCharge: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []


class AccessPlatformFilter(_Stem, BaseFilter):
    """Class for defining filter rules for access platform items."""

    entityType: Annotated[
        Literal["AccessPlatformFilter"], Field(alias="$type", frozen=True)
    ] = "AccessPlatformFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



class Joker:  # noqa: PLW1641
    """Testing utility that pretends to be equal to anything.

    Useful for tests that assert partially predictable dictionaries.

    Example:
        assert value == {"predictable": 42, "timestamp": Joker()}
    """

    _repr_cache: str | None = None

    def __eq__(self, obj: object) -> bool:
        """Pretend to be equal to any other object."""
        self._repr_cache = repr(obj)
        return True

    def __repr__(self) -> str:
        """Pretend to have the same representation as the last compared-to object."""
        # This hack allows pytest to print dict diffs without outlining jokers as diffs.
        if self._repr_cache:
            return self._repr_cache
        return super().__repr__()



"""Helper extractor to get metadata primary sources.

It represents the original source of metadata that all the data in MEx will attach to.
For example `confluence-vvt` primary source means: data extracted from `confluence-vvt`
x-system will be attached to this primary source.

Common use cases
----------------

- extract info of a particular primary source to attach extracted metadata to

Configuration
-------------

To configure primary_source extractor, set `primary_sources_path` in settings to point
to `primary-sources.json` in `mex-assets` repository. A sample primary sources file is
also included in `mex-extractors` at
`assets/raw-data/primary-sources/primary-sources.json` for testing purposes.

Extracting primary sources
--------------------------

Use `extract_seed_primary_sources` in `primary_source.extract` function to extract all
primary sources. This function will yield all the primary sources available in
`primary_sources.json` source file.

Transforming primary sources
----------------------------

Use `transform_seed_primary_sources_to_extracted_primary_sources` in
`primary_sources.transform` to get `ExtractedPrimarySource`. This function will yield
all the primary sources, which is often not required.

So to filter out only the required x-system primary sources use
`get_primary_sources_by_name` in `primary_sources.transform`. This function needs
Iterable from step-1 and names of the required x-systems. For example by passing names
as `mex`, `ldap`, `confluence-vvt` will return `ExtractedPrimarySource` of these
x-systems.
"""



from collections.abc import Generator
from typing import Any, TypeVar
from urllib.parse import urljoin

from mex.common.connector import HTTPConnector
from mex.common.identity.models import Identity
from mex.common.models import (
    AnyExtractedModel,
    AnyMergedModel,
    AnyPreviewModel,
    AnyRuleSetRequest,
    AnyRuleSetResponse,
    ExtractedContactPoint,
    ExtractedModelTypeAdapter,
    ExtractedOrganization,
    ExtractedPerson,
    ItemsContainer,
    MergedModelTypeAdapter,
    PaginatedItemsContainer,
    PreviewModelTypeAdapter,
    RuleSetResponseTypeAdapter,
)
from mex.common.models.person import MergedPerson
from mex.common.settings import BaseSettings
from mex.common.types import Identifier, MergedPrimarySourceIdentifier

_IngestibleModelT = TypeVar(
    "_IngestibleModelT", bound=AnyExtractedModel | AnyRuleSetResponse
)


class BackendApiConnector(HTTPConnector):
    """Connector class to handle interaction with the Backend API."""

    API_VERSION = "v0"

    def _check_availability(self) -> None:
        """Send a GET request to verify the API is available."""
        self.request("GET", "_system/check")

    def _set_authentication(self) -> None:
        """Set the backend API key to all session headers."""
        settings = BaseSettings.get()
        self.session.headers["X-API-Key"] = settings.backend_api_key.get_secret_value()

    def _set_url(self) -> None:
        """Set the backend api url with the version path."""
        settings = BaseSettings.get()
        self.url = urljoin(str(settings.backend_api_url), self.API_VERSION)

    def fetch_extracted_items(  # noqa: PLR0913
        self,
        *,
        query_string: str | None = None,
        stable_target_id: str | None = None,
        entity_type: list[str] | None = None,
        referenced_identifier: list[str] | None = None,
        reference_field: str | None = None,
        skip: int = 0,
        limit: int = 100,
    ) -> PaginatedItemsContainer[AnyExtractedModel]:
        """Fetch extracted items that match the given set of filters.

        Args:
            query_string: Full-text search query
            stable_target_id: The item's stableTargetId
            entity_type: The item's entityType
            referenced_identifier: Merged item identifiers filter
            reference_field: Field name to filter for
            skip: How many items to skip for pagination
            limit: How many items to return in one page

        Raises:
            HTTPError: If search was not accepted, crashes or times out

        Returns:
            One page of extracted items and the total count that was matched
        """
        response = self.request(
            method="GET",
            endpoint="extracted-item",
            params={
                "q": query_string,
                "stableTargetId": stable_target_id,
                "entityType": entity_type,
                "referencedIdentifier": referenced_identifier,
                "referenceField": reference_field,
                "skip": str(skip),
                "limit": str(limit),
            },
        )
        return PaginatedItemsContainer[AnyExtractedModel].model_validate(response)

    def get_extracted_item(
        self,
        identifier: str,
    ) -> AnyExtractedModel:
        """Return one extracted item for the given `identifier`.

        Args:
            identifier: The extracted item's identifier

        Raises:
            HTTPError: If no extracted item was found

        Returns:
            A single extracted item
        """
        response = self.request(
            method="GET",
            endpoint=f"extracted-item/{identifier}",
        )
        return ExtractedModelTypeAdapter.validate_python(response)

    def fetch_merged_items(  # noqa: PLR0913
        self,
        *,
        query_string: str | None = None,
        identifier: str | None = None,
        entity_type: list[str] | None = None,
        referenced_identifier: list[str] | None = None,
        reference_field: str | None = None,
        skip: int = 0,
        limit: int = 100,
    ) -> PaginatedItemsContainer[AnyMergedModel]:
        """Fetch merged items that match the given set of filters.

        Args:
            query_string: Full-text search query
            identifier: Merged item identifier filter
            entity_type: The items' entityType
            referenced_identifier: Merged item identifiers filter
            reference_field: Field name to filter for
            skip: How many items to skip for pagination
            limit: How many items to return in one page

        Raises:
            HTTPError: If search was not accepted, crashes or times out

        Returns:
            One page of merged items and the total count that was matched
        """
        response = self.request(
            method="GET",
            endpoint="merged-item",
            params={
                "q": query_string,
                "identifier": identifier,
                "entityType": entity_type,
                "referencedIdentifier": referenced_identifier,
                "referenceField": reference_field,
                "skip": str(skip),
                "limit": str(limit),
            },
        )
        return PaginatedItemsContainer[AnyMergedModel].model_validate(response)

    def fetch_all_merged_items(
        self,
        *,
        query_string: str | None = None,
        identifier: str | None = None,
        entity_type: list[str] | None = None,
        referenced_identifier: list[str] | None = None,
        reference_field: str | None = None,
    ) -> Generator[AnyMergedModel, None, None]:
        """Fetch all merged items that match the given set of filters.

        Args:
            query_string: Full-text search query
            identifier: Merged item identifier filter
            entity_type: The items' entityType
            referenced_identifier: Merged item identifiers filter
            reference_field: Field name to filter for

        Raises:
            HTTPError: If search was not accepted, crashes or times out

        Returns:
            Generator for all merged items that match the filters
        """
        response = self.fetch_merged_items(
            query_string=query_string,
            identifier=identifier,
            entity_type=entity_type,
            referenced_identifier=referenced_identifier,
            reference_field=reference_field,
            skip=0,
            limit=1,
        )
        total_item_number = response.total
        item_number_limit = 100  # 100 is the maximum possible number per get-request
        for item_counter in range(0, total_item_number, item_number_limit):
            response = self.fetch_merged_items(
                query_string=query_string,
                identifier=identifier,
                entity_type=entity_type,
                referenced_identifier=referenced_identifier,
                reference_field=reference_field,
                skip=item_counter,
                limit=item_number_limit,
            )
            yield from response.items

    def get_merged_item(
        self,
        identifier: str,
    ) -> AnyMergedModel:
        """Return one merged item for the given `identifier`.

        Args:
            identifier: The merged item's identifier

        Raises:
            HTTPError: If no merged item was found

        Returns:
            A single merged item
        """
        response = self.request(
            method="GET",
            endpoint=f"merged-item/{identifier}",
        )
        return MergedModelTypeAdapter.validate_python(response)

    def preview_merged_item(
        self,
        stable_target_id: str,
        rule_set: AnyRuleSetRequest,
    ) -> AnyPreviewModel:
        """Return a preview for merging the given rule-set with stored extracted items.

        Args:
            stable_target_id: The extracted items' `stableTargetId`
            rule_set: A rule-set to use for previewing

        Raises:
            HTTPError: If preview produces errors, crashes or times out

        Returns:
            A single merged item
        """
        response = self.request(
            method="POST",
            endpoint=f"preview-item/{stable_target_id}",
            payload=rule_set,
        )
        return PreviewModelTypeAdapter.validate_python(response)

    def fetch_preview_items(  # noqa: PLR0913
        self,
        *,
        query_string: str | None = None,
        identifier: str | None = None,
        entity_type: list[str] | None = None,
        referenced_identifier: list[str] | None = None,
        reference_field: str | None = None,
        skip: int = 0,
        limit: int = 100,
    ) -> PaginatedItemsContainer[AnyPreviewModel]:
        """Fetch merged item previews that match the given set of filters.

        Args:
            query_string: Full-text search query
            identifier: Merged item identifier filter
            entity_type: The items' entityType
            referenced_identifier: Merged item identifiers filter
            reference_field: Field name to filter for
            skip: How many items to skip for pagination
            limit: How many items to return in one page

        Raises:
            HTTPError: If search was not accepted, crashes or times out

        Returns:
            One page of preview items and the total count that was matched
        """
        response = self.request(
            method="GET",
            endpoint="preview-item",
            params={
                "q": query_string,
                "identifier": identifier,
                "entityType": entity_type,
                "referencedIdentifier": referenced_identifier,
                "referenceField": reference_field,
                "skip": str(skip),
                "limit": str(limit),
            },
        )
        return PaginatedItemsContainer[AnyPreviewModel].model_validate(response)

    def create_rule_set(
        self,
        rule_set: AnyRuleSetRequest,
    ) -> AnyRuleSetResponse:
        """Create a new rule set.

        Args:
            rule_set: New rule-set to create

        Raises:
            HTTPError: If the rule-set did not validate

        Returns:
            The newly created rule-set
        """
        response = self.request(method="POST", endpoint="rule-set", payload=rule_set)
        return RuleSetResponseTypeAdapter.validate_python(response)

    def get_rule_set(
        self,
        stable_target_id: str,
    ) -> AnyRuleSetResponse:
        """Return a triple of rules for the given `stableTargetId`.

        Args:
            stable_target_id: The merged item's identifier

        Raises:
            HTTPError: If no rule-set was found

        Returns:
            A set of three rules
        """
        response = self.request(
            method="GET",
            endpoint=f"rule-set/{stable_target_id}",
        )
        return RuleSetResponseTypeAdapter.validate_python(response)

    def update_rule_set(
        self, stable_target_id: str, rule_set: AnyRuleSetRequest
    ) -> AnyRuleSetResponse:
        """Update an existing rule set.

        Args:
            stable_target_id: The merged item's identifier
            rule_set: The new rule-set contents

        Raises:
            HTTPError: If no rule-set was found

        Returns:
            A set of three rules
        """
        response = self.request(
            method="PUT", endpoint=f"rule-set/{stable_target_id}", payload=rule_set
        )
        return RuleSetResponseTypeAdapter.validate_python(response)

    def search_organization_in_wikidata(
        self,
        q: str,
        offset: int = 0,
        limit: int = 10,
    ) -> PaginatedItemsContainer[ExtractedOrganization]:
        """Search for organizations in wikidata.

        Args:
            q: Wikidata item ID or full URL
            offset: The starting index for pagination
            limit: The maximum number of results to return

        Returns:
            Paginated list of ExtractedOrganizations
        """
        response = self.request(
            method="GET",
            endpoint="wikidata",
            params={"q": q, "offset": str(offset), "limit": str(limit)},
        )
        return PaginatedItemsContainer[ExtractedOrganization].model_validate(response)

    def search_person_or_contact_point_in_ldap(
        self,
        q: str,
        offset: int = 0,
        limit: int = 10,
    ) -> PaginatedItemsContainer[ExtractedPerson | ExtractedContactPoint]:
        """Search for persons or contact points in LDAP.

        Args:
            q: The name of the person or contact point
            offset: The starting index for pagination
            limit: The maximum number of results to return

        Returns:
            Paginated list of ExtractedPersons and ExtractedContactPoints
        """
        response = self.request(
            method="GET",
            endpoint="ldap",
            params={"q": q, "offset": str(offset), "limit": str(limit)},
        )
        return PaginatedItemsContainer[
            ExtractedPerson | ExtractedContactPoint
        ].model_validate(response)

    def search_person_in_orcid(
        self,
        q: str,
        offset: int = 0,
        limit: int = 10,
    ) -> PaginatedItemsContainer[ExtractedPerson]:
        """Search for persons in orcid.

        Args:
            q: The name of the person to be searched
            offset: The starting index for pagination
            limit: The maximum number of results to return

        Returns:
            Paginated list of ExtractedPersons
        """
        response = self.request(
            method="GET",
            endpoint="orcid",
            params={"q": q, "offset": str(offset), "limit": str(limit)},
        )
        return PaginatedItemsContainer[ExtractedPerson].model_validate(response)

    def assign_identity(
        self,
        had_primary_source: MergedPrimarySourceIdentifier,
        identifier_in_primary_source: str,
    ) -> Identity:
        """Find an Identity in a database or assign a new one."""
        response = self.request(
            "POST",
            "identity",
            {
                "hadPrimarySource": had_primary_source,
                "identifierInPrimarySource": identifier_in_primary_source,
            },
        )
        return Identity.model_validate(response)

    def fetch_identities(
        self,
        had_primary_source: Identifier | None = None,
        identifier_in_primary_source: str | None = None,
        stable_target_id: Identifier | None = None,
    ) -> ItemsContainer[Identity]:
        """Find Identity instances matching the given filters.

        Either provide `stableTargetId` or `hadPrimarySource`
        and `identifierInPrimarySource` together to get a unique result.
        """
        connector = BackendApiConnector.get()
        response = connector.request(
            "GET",
            "identity",
            params={
                "hadPrimarySource": had_primary_source,
                "identifierInPrimarySource": identifier_in_primary_source,
                "stableTargetId": stable_target_id,
            },
        )
        return ItemsContainer[Identity].model_validate(response)

    def ingest(
        self,
        ingestible_models: list[_IngestibleModelT],
        **kwargs: Any,  # noqa: ANN401
    ) -> None:
        """Post extracted models or rule-sets to the backend in bulk.

        Args:
            ingestible_models: Extracted models or rule-sets to ingest
            kwargs: Further keyword arguments passed to `requests`

        Raises:
            HTTPError: If post was not accepted, crashes or times out
        """
        self.request(
            method="POST",
            endpoint="ingest",
            payload=ItemsContainer[_IngestibleModelT](items=ingestible_models),
            **kwargs,
        )

    def merged_person_from_login(self) -> MergedPerson:
        """Fetch a MergedPerson by their login username.

        The username is automatically extracted from authentication headers
        by the backend endpoint dependency.

        Raises:
            HTTPError: If authentication fails or user is not authorized

        Returns:
            MergedPerson: The merged person corresponding to the authenticated user
        """
        response = self.request(
            method="POST",
            endpoint="merged-person-from-login",
        )
        return MergedPerson.model_validate(response)



import re
from typing import Any, cast
from urllib.parse import urlsplit

import backoff
from ldap3 import AUTO_BIND_NO_TLS, Connection, Server
from ldap3.core.exceptions import LDAPExceptionError, LDAPSocketSendError

from mex.common.connector import BaseConnector
from mex.common.exceptions import (
    EmptySearchResultError,
    FoundMoreThanOneError,
    MExError,
)
from mex.common.ldap.models import (
    LDAP_MODEL_CLASSES,
    AnyLDAPActor,
    AnyLDAPActorsTypeAdapter,
    LDAPFunctionalAccount,
    LDAPFunctionalAccountsTypeAdapter,
    LDAPPerson,
    LDAPPersonsTypeAdapter,
)
from mex.common.logging import logger
from mex.common.settings import BaseSettings


class LDAPConnector(BaseConnector):
    """Connector class to handle credentials and querying of LDAP."""

    def __init__(self) -> None:
        """Create a new LDAP connection."""
        settings = BaseSettings.get()
        self._search_base = settings.ldap_search_base
        self._connection = self._setup_connection()

    def _setup_connection(self) -> Connection:
        """Set up a new LDAP connection."""
        settings = BaseSettings.get()
        url = urlsplit(settings.ldap_url.get_secret_value())
        host = str(url.hostname)
        port = int(url.port) if url.port else None
        server = Server(host, port, use_ssl=True)
        connection = Connection(
            server,
            user=url.username,
            password=url.password,
            auto_bind=AUTO_BIND_NO_TLS,
            read_only=True,
        )
        connection.__enter__()
        try:
            connection.server.check_availability()
        except LDAPExceptionError as error:
            msg = f"LDAP service not available at url: {host}:{port}"
            raise MExError(msg) from error
        return connection

    def close(self) -> None:
        """Close the connector's underlying LDAP connection."""
        if self._connection:
            self._connection.__exit__(None, None, None)

    def reconnect(self) -> None:
        """Close current ldap connection and initiate a new one."""
        self.close()
        self._connection = self._setup_connection()

    @backoff.on_exception(
        wait_gen=backoff.fibo,
        exception=(LDAPSocketSendError,),
        max_tries=2,
        logger=logger,
        on_backoff=lambda details: cast(
            "LDAPConnector", details["args"][0]
        ).reconnect(),
    )
    def _fetch(self, search_filter: str, limit: int) -> list[dict[str, Any]]:
        """Fetch all items that match the given filters.

        Args:
            search_filter: LDAP search query
            limit: How many items to return

        Returns:
            List of raw ldap items
        """
        response = self._connection.extend.standard.paged_search(
            search_base=self._search_base,
            search_filter=search_filter.strip(),
            attributes=tuple(
                sorted({f for m in LDAP_MODEL_CLASSES for f in m.model_fields})
            ),
            size_limit=limit,
        )
        return [
            attributes for item in response if (attributes := item.get("attributes"))
        ]

    @staticmethod
    def _sanitize(value: str) -> str:
        """Sanitize a query value by removing invalid chars."""
        return re.sub(r"[^a-zA-ZÀ-ÿ0-9\*\.\-_@ ]+", "*", value)

    def get_persons_or_functional_accounts(
        self,
        *,
        query: str = "*",
        limit: int = 10,
    ) -> list[AnyLDAPActor]:
        """Get LDAP persons or functional accounts.

        Args:
            query: Display name of person or email of functional account
            limit: How many items to return

        Returns:
            List of LDAP persons and/or functional accounts
        """
        search_filter = f"""
        (|
            (&
                (objectCategory=Person)
                (sAMAccountName=*)
                (employeeID=*)
                (displayName={self._sanitize(query)})
            )
            (&
                (objectCategory=Person)
                (OU=Funktion)
                (mail={self._sanitize(query)})
            )
        )
        """
        raw_items = self._fetch(search_filter, limit)
        return AnyLDAPActorsTypeAdapter.validate_python(raw_items)

    def get_functional_accounts(
        self,
        *,
        mail: str = "*",
        limit: int = 10,
    ) -> list[LDAPFunctionalAccount]:
        """Get LDAP functional accounts that match provided filters.

        Some projects/resources declare functional mailboxes as their contact.

        Args:
            mail: Email address of the functional account
            limit: How many items to return

        Returns:
            List of LDAP functional accounts
        """
        search_filter = f"""
        (&
            (objectCategory=Person)
            (OU=Funktion)
            (mail={self._sanitize(mail)})
        )
        """
        raw_items = self._fetch(search_filter, limit)
        return LDAPFunctionalAccountsTypeAdapter.validate_python(raw_items)

    def get_persons(  # noqa: PLR0913
        self,
        *,
        display_name: str = "*",
        employee_id: str = "*",
        given_name: str = "*",
        mail: str = "*",
        object_guid: str = "*",
        sam_account_name: str = "*",
        surname: str = "*",
        limit: int = 10,
    ) -> list[LDAPPerson]:
        """Get LDAP persons that match the provided filters.

        An LDAP person's objectGUIDs is stable across name changes, whereas name based
        person identifiers of the schema SurnameF are not stable.

        Args:
            display_name: Display name of the person
            employee_id: Employee identifier
            given_name: Given name of a person, defaults to non-null
            mail: Email address, defaults to non-null
            object_guid: Internal LDAP identifier
            sam_account_name: Account name
            surname: Surname of a person, defaults to non-null
            limit: How many items to return

        Returns:
            List of LDAP persons
        """
        search_filter = f"""
        (&
            (objectCategory=Person)
            (displayName={self._sanitize(display_name)})
            (employeeID={self._sanitize(employee_id)})
            (givenName={self._sanitize(given_name)})
            (mail={self._sanitize(mail)})
            (objectGUID={self._sanitize(object_guid)})
            (sAMAccountName={self._sanitize(sam_account_name)})
            (sn={self._sanitize(surname)})
        )
        """
        raw_items = self._fetch(search_filter, limit)
        return LDAPPersonsTypeAdapter.validate_python(raw_items)

    def get_functional_account(
        self,
        *,
        mail: str = "*",
    ) -> LDAPFunctionalAccount:
        """Get a single LDAP functional account for the given mail address.

        Args:
            mail: Email address of the functional account

        Raises:
            EmptySearchResultError: If no LDAP entry matching filters was found
            FoundMoreThanOneError: If more than one LDAP entry was found

        Returns:
            Single LDAP functional account
        """
        functional_accounts = self.get_functional_accounts(mail=mail, limit=2)
        if not functional_accounts:
            msg = f"Cannot find AD functional account for filters mail: {mail}"
            raise EmptySearchResultError(msg)
        if len(functional_accounts) > 1:
            msg = f"Found multiple AD functional accounts for mail: {mail}"
            raise FoundMoreThanOneError(msg)
        return functional_accounts[0]

    def get_person(  # noqa: PLR0913
        self,
        *,
        display_name: str = "*",
        employee_id: str = "*",
        given_name: str = "*",
        mail: str = "*",
        object_guid: str = "*",
        sam_account_name: str = "*",
        surname: str = "*",
    ) -> LDAPPerson:
        """Get a single LDAP person for the given filters.

        Args:
            display_name: Display name of the person
            employee_id: Employee identifier
            given_name: Given name of a person, defaults to non-null
            mail: Email address, defaults to non-null
            object_guid: Internal LDAP identifier
            sam_account_name: Account name
            surname: Surname of a person, defaults to non-null

        Raises:
            EmptySearchResultError: If no LDAP entry matching filters was found
            FoundMoreThanOneError: If more than one LDAP entry was found

        Returns:
            Single LDAP person matching the filters
        """
        persons = self.get_persons(
            employee_id=employee_id,
            given_name=given_name,
            mail=mail,
            object_guid=object_guid,
            sam_account_name=sam_account_name,
            surname=surname,
            display_name=display_name,
            limit=2,
        )
        if not persons:
            msg = "Cannot find AD person for filters"
            raise EmptySearchResultError(msg)
        if len(persons) > 1:
            msg = "Found multiple AD persons for filters"
            raise FoundMoreThanOneError(msg)
        return persons[0]



from collections.abc import Iterable

from mex.common.logging import logger
from mex.common.models import (
    ExtractedOrganizationalUnit,
)
from mex.common.organigram.models import OrganigramUnit
from mex.common.types import (
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPrimarySourceIdentifier,
)


def transform_organigram_unit_to_extracted_organizational_unit(
    organigram_unit: OrganigramUnit,
    primary_source_id: MergedPrimarySourceIdentifier,
    rki_organization_id: MergedOrganizationIdentifier,
) -> ExtractedOrganizationalUnit:
    """Transform an organigram unit into an ExtractedOrganizationalUnit.

    Args:
        organigram_unit: Iterable of organigram units coming from the JSON file
        primary_source_id: Primary source identifier for organigram
        rki_organization_id: identifier of RKI organization to which the unit belongs

    Returns:
        ExtractedOrganizationalUnit
    """
    return ExtractedOrganizationalUnit(
        identifierInPrimarySource=organigram_unit.identifier,
        hadPrimarySource=primary_source_id,
        alternativeName=organigram_unit.alternativeName,
        unitOf=[rki_organization_id],
        email=organigram_unit.email,
        name=organigram_unit.name,
        shortName=organigram_unit.shortName,
        website=[organigram_unit.website] if organigram_unit.website else [],
    )


def transform_organigram_units_to_organizational_units(
    organigram_units: Iterable[OrganigramUnit],
    primary_source_id: MergedPrimarySourceIdentifier,
    rki_organization_id: MergedOrganizationIdentifier,
) -> list[ExtractedOrganizationalUnit]:
    """Transform organigram units into ExtractedOrganizationalUnits.

    Beware that the order of the output is not necessarily the order of the input.

    Args:
        organigram_units: Iterable of organigram units coming from the JSON file
        primary_source_id: Primary source for organigram
        rki_organization_id: identifier of RKI organization to which the units belong

    Returns:
        List of ExtractedOrganizationalUnit
    """
    extracted_unit_by_id_in_primary_source: dict[str, ExtractedOrganizationalUnit] = {}
    parent_id_in_primary_source_by_id_in_primary_source: dict[str, str] = {}

    for unit in organigram_units:
        extracted_unit = transform_organigram_unit_to_extracted_organizational_unit(
            unit, primary_source_id, rki_organization_id
        )
        extracted_unit_by_id_in_primary_source[unit.identifier] = extracted_unit
        if parent_identifier_in_primary_source := unit.parentUnit:
            parent_id_in_primary_source_by_id_in_primary_source[unit.identifier] = (
                parent_identifier_in_primary_source
            )

    for extracted_unit in extracted_unit_by_id_in_primary_source.values():
        identifier_in_primary_source = extracted_unit.identifierInPrimarySource
        if (  # noqa: SIM102
            parent_identifier_in_primary_source
            := parent_id_in_primary_source_by_id_in_primary_source.get(
                identifier_in_primary_source
            )
        ):
            if parent_unit := extracted_unit_by_id_in_primary_source.get(
                parent_identifier_in_primary_source
            ):
                extracted_unit.parentUnit = MergedOrganizationalUnitIdentifier(
                    parent_unit.stableTargetId
                )
    logger.info(
        "transformed %s organizational units",
        len(extracted_unit_by_id_in_primary_source),
    )
    return list(extracted_unit_by_id_in_primary_source.values())



from typing import TYPE_CHECKING, Final

from mex.common.types import IdentityProvider

if TYPE_CHECKING:
    from mex.common.identity.base import BaseProvider

_PROVIDER_REGISTRY: Final[dict[IdentityProvider, type["BaseProvider"]]] = {}


def register_provider(
    key: IdentityProvider, provider_cls: type["BaseProvider"]
) -> None:
    """Register an implementation of an identity provider to a settings key.

    Args:
        key: Possible value of `BaseSettings.identity_provider`
        provider_cls: Implementation of an identity provider

    Raises:
        RuntimeError: When the `key` is already registered
    """
    if key in _PROVIDER_REGISTRY:
        msg = f"Already registered identity provider: {key}"
        raise RuntimeError(msg)
    _PROVIDER_REGISTRY[key] = provider_cls


def get_provider() -> "BaseProvider":
    """Get an instance of the identity provider as configured by `identity_provider`.

    Raises:
        RuntimeError: When the configured provider is not registered

    Returns:
        An instance of a subclass of `BaseProvider`
    """
    # break import cycle, sigh
    from mex.common.settings import BaseSettings  # noqa: PLC0415

    settings = BaseSettings.get()
    if settings.identity_provider in _PROVIDER_REGISTRY:
        provider_cls = _PROVIDER_REGISTRY[settings.identity_provider]
        return provider_cls.get()
    msg = f"Identity provider not implemented: {settings.identity_provider}"
    raise RuntimeError(msg)



import logging
import logging.config
from collections.abc import Callable, Generator
from functools import wraps
from typing import Any, ParamSpec, TypeVar

import click

_YieldT = TypeVar("_YieldT")

LOGGING_CONFIG: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": f"{click.style('%(asctime)s', fg='bright_yellow')}"
            " - mex - %(message)s",
        }
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "mex": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": True,
        }
    },
}
logging.config.dictConfig(LOGGING_CONFIG)
logger = logging.getLogger("mex")
echo = logger.info  # `echo` is deprecated, use the logger directly instead

P = ParamSpec("P")


def watch(
    log_interval: int = 10000,
) -> Callable[
    [Callable[P, Generator[_YieldT, None, None]]],
    Callable[P, Generator[_YieldT, None, None]],
]:
    """Watch the output of a generator function and log the yielded items.

    Args:
        func: Generator function that yields strings, models or exceptions
            (It will use the objects `__str__()` method to print it out.)
        log_interval: integer determining the interval length between loggings

    Returns:
        Decorated function that logs all yielded items
    """

    def decorator(
        func: Callable[P, Generator[_YieldT, None, None]],
    ) -> Callable[P, Generator[_YieldT, None, None]]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Generator[_YieldT, None, None]:  # noqa: ANN401
            for i, item in enumerate(func(*args, **kwargs)):
                if i % log_interval == 0:
                    logger.info("%s - %s - %s", i, func.__name__, item)
                yield item

        return wrapper

    return decorator



from mex.common.models import BaseModel
from mex.common.types import Link, Text


class OrganigramUnit(BaseModel):
    """Organizational units in the format of the organigram JSON file."""

    shortName: list[Text]
    email: list[str] = []
    identifier: str
    name: list[Text]
    alternativeName: list[Text] = []
    parentUnit: None | str = None
    website: None | Link = None



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedPersonIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
)

EmailStr = Annotated[
    str,
    Field(
        examples=["info@rki.de"],
        pattern="^[^@ \\t\\r\\n]+@[^@ \\t\\r\\n]+\\.[^@ \\t\\r\\n]+$",
        json_schema_extra={"format": "email"},
    ),
]
FamilyNameStr = Annotated[
    str,
    Field(
        examples=["Patapoutian", "Skłodowska-Curie", "Muta Maathai"],
    ),
]
FullNameStr = Annotated[
    str,
    Field(
        examples=["Juturna Felicitás", "M. Berhanu", "Keone Seong-Hyeon"],
    ),
]
GivenNameStr = Annotated[
    str,
    Field(
        examples=["Romāns", "Marie Salomea", "May-Britt"],
    ),
]
IsniIdStr = Annotated[
    str,
    Field(
        pattern="^https://isni\\.org/isni/[X0-9]{16}$",
        examples=[
            "https://isni.org/isni/0000000019240398",
            "https://isni.org/isni/0000000453907343",
            "https://isni.org/isni/0000000038086111",
        ],
        json_schema_extra={"format": "uri"},
    ),
]
OrcidIdStr = Annotated[
    str,
    Field(
        pattern="^https://orcid\\.org/[-X0-9]{9,21}$",
        examples=[
            "https://orcid.org/0000-0002-9079-593X",
            "https://orcid.org/0000-0003-2300-3928",
            "https://orcid.org/0000-0002-1335-4022",
        ],
        json_schema_extra={"format": "uri"},
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Person"], Field(frozen=True)]] = "Person"


class _OptionalLists(_Stem):
    affiliation: Annotated[
        list[MergedOrganizationIdentifier],
        Field(
            description="An organization that the described person is affiliated with.",
            json_schema_extra={
                "sameAs": [
                    "https://schema.org/affiliation",
                    "http://www.wikidata.org/entity/P1416",
                ]
            },
        ),
    ] = []
    email: Annotated[
        list[EmailStr],
        Field(
            description="The email address through which the person can be contacted.",
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/2006/vcard/ns#hasEmail",
                    "https://schema.org/email",
                ]
            },
        ),
    ] = []
    familyName: Annotated[
        list[FamilyNameStr],
        Field(
            description="The name inherited from the family.",
            json_schema_extra={
                "sameAs": [
                    "http://xmlns.com/foaf/0.1/familyName",
                    "https://schema.org/familyName",
                ]
            },
        ),
    ] = []
    fullName: Annotated[
        list[FullNameStr],
        Field(
            description=(
                "The full name of a person. Also used if the naming schema "
                "(given name and family name) does not apply to the name."
            ),
            json_schema_extra={"sameAs": ["http://xmlns.com/foaf/0.1/name"]},
        ),
    ] = []
    givenName: Annotated[
        list[GivenNameStr],
        Field(
            description="The name given to the person e.g. by their parents.",
            json_schema_extra={
                "sameAs": [
                    "http://xmlns.com/foaf/0.1/givenName",
                    "https://schema.org/givenName",
                ]
            },
        ),
    ] = []
    isniId: Annotated[
        list[IsniIdStr],
        Field(
            description=(
                "The ISNI (International Standard Name Identifier) of the person."
            ),
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P213"]},
        ),
    ] = []
    memberOf: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            description="Organizational unit at RKI the person is associated with.",
            json_schema_extra={
                "sameAs": [
                    "http://www.cidoc-crm.org/cidoc-crm/P107i_is_current_or_former_member_of"
                ]
            },
        ),
    ] = []
    orcidId: Annotated[
        list[OrcidIdStr],
        Field(
            description="Identifier of a person from the ORCID authority file.",
            json_schema_extra={"sameAs": ["http://www.wikidata.org/entity/P496"]},
        ),
    ] = []


class BasePerson(
    _OptionalLists,
    json_schema_extra={
        "description": (
            "A person ([FOAF, 2004-05-01](http://xmlns.com/foaf/0.1/)). This class "
            "comprises real persons who live or are assumed to have lived ([CIDOC CRM, "
            "version 7.1.1](https://cidoc-crm.org/html/cidoc_crm_v7.1.1.html))."
        ),
        "sameAs": [
            "http://www.cidoc-crm.org/cidoc-crm/E21_Person",
            "http://xmlns.com/foaf/0.1/Person",
            "http://www.w3.org/2006/vcard/ns#Individual",
        ],
    },
):
    """All fields for a valid person except for provenance."""


class ExtractedPerson(BasePerson, ExtractedData):
    """An automatically extracted metadata set describing a person."""

    entityType: Annotated[
        Literal["ExtractedPerson"], Field(alias="$type", frozen=True)
    ] = "ExtractedPerson"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedPersonIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedPersonIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedPersonIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedPersonIdentifier)


class MergedPerson(BasePerson, MergedItem):
    """The result of merging all extracted items and rules for a person."""

    entityType: Annotated[
        Literal["MergedPerson"], Field(alias="$type", frozen=True)
    ] = "MergedPerson"
    identifier: Annotated[
        MergedPersonIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedPersonIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewPerson(_OptionalLists, PreviewItem):
    """Preview for merging all extracted items and rules for a person."""

    entityType: Annotated[
        Literal["PreviewPerson"], Field(alias="$type", frozen=True)
    ] = "PreviewPerson"
    identifier: Annotated[
        MergedPersonIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedPersonIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditivePerson(_OptionalLists, AdditiveRule):
    """Rule to add values to merged person items."""

    entityType: Annotated[
        Literal["AdditivePerson"], Field(alias="$type", frozen=True)
    ] = "AdditivePerson"
    supersededBy: Annotated[
        MergedPersonIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractivePerson(_OptionalLists, SubtractiveRule):
    """Rule to subtract values from merged person items."""

    entityType: Annotated[
        Literal["SubtractivePerson"], Field(alias="$type", frozen=True)
    ] = "SubtractivePerson"


class PreventivePerson(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged person items."""

    entityType: Annotated[
        Literal["PreventivePerson"], Field(alias="$type", frozen=True)
    ] = "PreventivePerson"
    affiliation: list[MergedPrimarySourceIdentifier] = []
    email: list[MergedPrimarySourceIdentifier] = []
    familyName: list[MergedPrimarySourceIdentifier] = []
    fullName: list[MergedPrimarySourceIdentifier] = []
    givenName: list[MergedPrimarySourceIdentifier] = []
    isniId: list[MergedPrimarySourceIdentifier] = []
    memberOf: list[MergedPrimarySourceIdentifier] = []
    orcidId: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for a person item."""

    additive: AdditivePerson = AdditivePerson()
    subtractive: SubtractivePerson = SubtractivePerson()
    preventive: PreventivePerson = PreventivePerson()


class PersonRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a person item."""

    entityType: Annotated[
        Literal["PersonRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "PersonRuleSetRequest"


class PersonRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a person item."""

    entityType: Annotated[
        Literal["PersonRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "PersonRuleSetResponse"
    stableTargetId: MergedPersonIdentifier


class PersonMapping(_Stem, BaseMapping):
    """Mapping for describing a person transformation."""

    entityType: Annotated[
        Literal["PersonMapping"], Field(alias="$type", frozen=True)
    ] = "PersonMapping"
    affiliation: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    email: list[MappingField[list[EmailStr]]] = []
    familyName: list[MappingField[list[FamilyNameStr]]] = []
    fullName: list[MappingField[list[FullNameStr]]] = []
    givenName: list[MappingField[list[GivenNameStr]]] = []
    isniId: list[MappingField[list[IsniIdStr]]] = []
    memberOf: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []
    orcidId: list[MappingField[list[OrcidIdStr]]] = []


class PersonFilter(_Stem, BaseFilter):
    """Class for defining filter rules for person items."""

    entityType: Annotated[
        Literal["PersonFilter"], Field(alias="$type", frozen=True)
    ] = "PersonFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



import json
from functools import lru_cache
from pathlib import Path

from mex.common.logging import logger
from mex.common.primary_source.models import SeedPrimarySource
from mex.common.settings import BaseSettings


@lru_cache(maxsize=1)
def extract_seed_primary_sources() -> list[SeedPrimarySource]:
    """Extract seed primary sources from the raw-data JSON file.

    Settings:
        primary_sources_path: Resolved path to the primary sources file

    Returns:
        List of seed primary sources
    """
    settings = BaseSettings.get()
    with Path(settings.primary_sources_path).open() as fh:
        raw_units = json.load(fh)
    logger.info("extracted %s seed primary sources", len(raw_units))
    return [SeedPrimarySource.model_validate(raw) for raw in raw_units]



from collections.abc import Generator, Iterable
from itertools import tee
from typing import Final

from mex.common.models import AnyExtractedModel, AnyMergedModel, AnyRuleSetResponse
from mex.common.settings import BaseSettings
from mex.common.sinks.backend_api import BackendApiSink
from mex.common.sinks.base import BaseSink
from mex.common.sinks.ndjson import NdjsonSink
from mex.common.types import Sink

_SINK_REGISTRY: Final[dict[Sink, type["BaseSink"]]] = {}


class _MultiSink(BaseSink):
    """Sink to load models to multiple sinks simultaneously."""

    # This class is private because it should only be acquired by calling `get_sink`.

    _sinks: list[BaseSink] = []

    def __init__(self) -> None:
        """Instantiate the multi sink singleton."""
        settings = BaseSettings.get()
        for sink in settings.sink:
            if sink in _SINK_REGISTRY:
                sink_cls = _SINK_REGISTRY[sink]
                self._sinks.append(sink_cls.get())
            else:
                msg = f"Sink function not implemented: {sink}"
                raise RuntimeError(msg)

    def close(self) -> None:
        """Close all underlying sinks."""
        for sink in self._sinks:
            sink.close()

    def load(
        self,
        items: Iterable[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse],
    ) -> Generator[AnyExtractedModel | AnyMergedModel | AnyRuleSetResponse, None, None]:
        """Load the given items to multiple sinks simultaneously."""
        for sink, model_gen in zip(
            self._sinks, tee(items, len(self._sinks)), strict=True
        ):
            yield from sink.load(model_gen)


def register_sink(key: Sink, sink_cls: type["BaseSink"]) -> None:
    """Register an implementation of a sink function to a settings key.

    Args:
        key: Possible value of `BaseSettings.sink`
        sink_cls: Implementation of the abstract sink class

    Raises:
        RuntimeError: When the `key` is already registered
    """
    if key in _SINK_REGISTRY:
        msg = f"Already registered sink function: {key}"
        raise RuntimeError(msg)
    _SINK_REGISTRY[key] = sink_cls


def get_sink() -> "BaseSink":
    """Get a sink function that serves all configured `sink` destinations.

    Raises:
        RuntimeError: When the configured sink is not registered

    Returns:
        A function that pours the models into all configured sinks
    """
    return _MultiSink.get()


# register the default providers shipped with mex-common
register_sink(Sink.BACKEND, BackendApiSink)
register_sink(Sink.NDJSON, NdjsonSink)



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedVariableIdentifier,
    MergedPrimarySourceIdentifier,
    MergedResourceIdentifier,
    MergedVariableGroupIdentifier,
    MergedVariableIdentifier,
    Text,
)

CodingSystemStr = Annotated[
    str,
    Field(
        examples=["SF-36 Version 1"],
    ),
]
DataTypeStr = Annotated[
    str,
    Field(
        examples=["integer", "string", "image", "int55", "number"],
    ),
]
ValueSetStr = Annotated[
    str,
    Field(
        examples=[
            "Ja, stark eingeschränkt",
            "Ja, etwas eingeschränkt",
            "Nein, überhaupt nicht eingeschränkt",
        ],
    ),
]
LabelText = Annotated[
    Text,
    Field(
        examples=[{"language": "de", "value": "Mehrere Treppenabsätze steigen"}],
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Variable"], Field(frozen=True)]] = "Variable"


class _OptionalLists(_Stem):
    belongsTo: Annotated[
        list[MergedVariableGroupIdentifier],
        Field(
            description=(
                "The variable group, the described variable is part of. Used to "
                "group variables together, depending on how they are structured in "
                "the primary source."
            ),
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/isPartOf"]},
        ),
    ] = []
    description: Annotated[
        list[Text],
        Field(
            description=(
                "A description of the variable. How the variable is defined in "
                "the primary source."
            ),
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/description"]},
        ),
    ] = []
    valueSet: Annotated[
        list[ValueSetStr],
        Field(description="A set of predefined values as given in the primary source."),
    ] = []


class _RequiredLists(_Stem):
    label: Annotated[
        list[LabelText],
        Field(
            description="The name of the variable.",
            min_length=1,
            json_schema_extra={
                "sameAs": [
                    "http://purl.org/dc/terms/title",
                    "http://www.w3.org/2000/01/rdf-schema#label",
                ]
            },
        ),
    ]
    usedIn: Annotated[
        list[MergedResourceIdentifier],
        Field(
            description="The resource, the variable is used in.",
            min_length=1,
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/isPartOf"]},
        ),
    ]


class _SparseLists(_Stem):
    label: Annotated[
        list[LabelText],
        Field(
            description="The name of the variable.",
            json_schema_extra={
                "sameAs": [
                    "http://purl.org/dc/terms/title",
                    "http://www.w3.org/2000/01/rdf-schema#label",
                ]
            },
        ),
    ] = []
    usedIn: Annotated[
        list[MergedResourceIdentifier],
        Field(
            description="The resource, the variable is used in.",
            json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/isPartOf"]},
        ),
    ] = []


class _OptionalValues(_Stem):
    codingSystem: Annotated[
        CodingSystemStr | None,
        Field(
            description=(
                "An established standard to which the described resource conforms "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/"
                "dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": [
                    "http://purl.org/dc/terms/conformsTo",
                    "https://schema.org/codingSystem",
                ]
            },
        ),
    ] = None
    dataType: Annotated[
        DataTypeStr | None,
        Field(
            description="The defined data type of the variable.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/csvw#datatype"]},
        ),
    ] = None


class _VariadicValues(_Stem):
    codingSystem: Annotated[
        list[CodingSystemStr],
        Field(
            description=(
                "An established standard to which the described resource conforms "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/"
                "dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": [
                    "http://purl.org/dc/terms/conformsTo",
                    "https://schema.org/codingSystem",
                ]
            },
        ),
    ] = []
    dataType: Annotated[
        list[DataTypeStr],
        Field(
            description="The defined data type of the variable.",
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/csvw#datatype"]},
        ),
    ] = []


class BaseVariable(
    _OptionalLists,
    _RequiredLists,
    _OptionalValues,
    json_schema_extra={
        "description": (
            "Variables are defined for the data-based evaluation of investigations "
            "(e.g. studies). A variable is characterized by its data type (e.g. "
            "integer, string, date) and value range. The variable can be either "
            "quantitative or qualitative, i.e. the value range can take numerical or "
            "categorical values."
        ),
    },
):
    """All fields for a valid variable except for provenance."""


class ExtractedVariable(BaseVariable, ExtractedData):
    """An automatically extracted metadata set describing a variable."""

    entityType: Annotated[
        Literal["ExtractedVariable"], Field(alias="$type", frozen=True)
    ] = "ExtractedVariable"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(  # noqa: D102
        self,
    ) -> Annotated[
        ExtractedVariableIdentifier,
        Field(
            description=(
                "An unambiguous reference to the resource within a given context. "
                "Persistent identifiers should be provided as HTTP URIs "
                "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
            ),
            json_schema_extra={
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"]
            },
        ),
    ]:
        return self._get_identifier(ExtractedVariableIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(  # noqa: D102, N802
        self,
    ) -> Annotated[
        MergedVariableIdentifier,
        Field(
            description=(
                "The identifier of the merged item that this extracted item belongs to."
            )
        ),
    ]:
        return self._get_stable_target_id(MergedVariableIdentifier)


class MergedVariable(BaseVariable, MergedItem):
    """The result of merging all extracted items and rules for a variable."""

    entityType: Annotated[
        Literal["MergedVariable"], Field(alias="$type", frozen=True)
    ] = "MergedVariable"
    identifier: Annotated[
        MergedVariableIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedVariableIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class PreviewVariable(_OptionalLists, _SparseLists, _OptionalValues, PreviewItem):
    """Preview for merging all extracted items and rules for a variable."""

    entityType: Annotated[
        Literal["PreviewVariable"], Field(alias="$type", frozen=True)
    ] = "PreviewVariable"
    identifier: Annotated[
        MergedVariableIdentifier,
        Field(
            json_schema_extra={
                "description": (
                    "An unambiguous reference to the resource within a given context. "
                    "Persistent identifiers should be provided as HTTP URIs "
                    "([DCT, 2020-01-20](http://dublincore.org/specifications/dublin-core/dcmi-terms/2020-01-20/))."
                ),
                "readOnly": True,
                "sameAs": ["http://purl.org/dc/elements/1.1/identifier"],
            },
            frozen=True,
        ),
    ]
    supersededBy: Annotated[
        MergedVariableIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class AdditiveVariable(_OptionalLists, _SparseLists, _OptionalValues, AdditiveRule):
    """Rule to add values to merged variable items."""

    entityType: Annotated[
        Literal["AdditiveVariable"], Field(alias="$type", frozen=True)
    ] = "AdditiveVariable"
    supersededBy: Annotated[
        MergedVariableIdentifier | None,
        Field(
            json_schema_extra={
                "description": (
                    "A merged item which is the preferred duplicate, because it "
                    "replaces, consolidates or otherwise makes the current merged item "
                    "obsolete."
                ),
            }
        ),
    ] = None


class SubtractiveVariable(
    _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
):
    """Rule to subtract values from merged variable items."""

    entityType: Annotated[
        Literal["SubtractiveVariable"], Field(alias="$type", frozen=True)
    ] = "SubtractiveVariable"


class PreventiveVariable(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged variable items."""

    entityType: Annotated[
        Literal["PreventiveVariable"], Field(alias="$type", frozen=True)
    ] = "PreventiveVariable"
    belongsTo: list[MergedPrimarySourceIdentifier] = []
    codingSystem: list[MergedPrimarySourceIdentifier] = []
    dataType: list[MergedPrimarySourceIdentifier] = []
    description: list[MergedPrimarySourceIdentifier] = []
    label: list[MergedPrimarySourceIdentifier] = []
    usedIn: list[MergedPrimarySourceIdentifier] = []
    valueSet: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    """Base class for sets of rules for a variable item."""

    additive: AdditiveVariable = AdditiveVariable()
    subtractive: SubtractiveVariable = SubtractiveVariable()
    preventive: PreventiveVariable = PreventiveVariable()


class VariableRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a variable item."""

    entityType: Annotated[
        Literal["VariableRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "VariableRuleSetRequest"


class VariableRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a variable item."""

    entityType: Annotated[
        Literal["VariableRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "VariableRuleSetResponse"
    stableTargetId: MergedVariableIdentifier


class VariableMapping(_Stem, BaseMapping):
    """Mapping for describing a variable transformation."""

    entityType: Annotated[
        Literal["VariableMapping"], Field(alias="$type", frozen=True)
    ] = "VariableMapping"
    codingSystem: list[MappingField[CodingSystemStr | None]] = []
    dataType: list[MappingField[DataTypeStr | None]] = []
    label: Annotated[list[MappingField[list[LabelText]]], Field(min_length=1)]
    usedIn: Annotated[
        list[MappingField[list[MergedResourceIdentifier]]], Field(min_length=1)
    ]
    belongsTo: list[MappingField[list[MergedVariableGroupIdentifier]]] = []
    description: list[MappingField[list[Text]]] = []
    valueSet: list[MappingField[list[ValueSetStr]]] = []


class VariableFilter(_Stem, BaseFilter):
    """Class for defining filter rules for variable items."""

    entityType: Annotated[
        Literal["VariableFilter"], Field(alias="$type", frozen=True)
    ] = "VariableFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from typing import Annotated

from pydantic import Field

from mex.common.models import BaseModel
from mex.common.types import Identifier, MergedPrimarySourceIdentifier


class Identity(BaseModel):
    """Model for identifier lookup."""

    identifier: Annotated[Identifier, Field(frozen=True)]
    hadPrimarySource: Annotated[MergedPrimarySourceIdentifier, Field(frozen=True)]
    identifierInPrimarySource: Annotated[str, Field(frozen=True)]
    stableTargetId: Annotated[Identifier, Field(frozen=True)]



import hashlib
import pickle
from collections.abc import MutableMapping
from typing import Any, Literal

from pydantic import BaseModel as PydanticBaseModel
from pydantic import ValidatorFunctionWrapHandler, model_validator
from pydantic.json_schema import DEFAULT_REF_TEMPLATE, JsonSchemaMode
from pydantic.json_schema import GenerateJsonSchema as PydanticJsonSchemaGenerator

from mex.common.models.base.schema import JsonSchemaGenerator
from mex.common.utils import (
    get_alias_lookup,
    get_all_fields,
    get_field_names_allowing_none,
    get_list_field_names,
)


class BaseModel(
    PydanticBaseModel,
    str_strip_whitespace=True,
    populate_by_name=True,
    extra="ignore",
    str_max_length=10**5,
    str_min_length=1,
    use_enum_values=True,
    validate_default=True,
    validate_assignment=True,
):
    """Common base class for all MEx model classes."""

    @classmethod
    def model_json_schema(
        cls,
        by_alias: bool = True,  # noqa: FBT001, FBT002
        ref_template: str = DEFAULT_REF_TEMPLATE,
        schema_generator: type[PydanticJsonSchemaGenerator] = JsonSchemaGenerator,
        mode: JsonSchemaMode = "validation",
        *,
        union_format: Literal["any_of", "primitive_type_array"] = "any_of",
    ) -> dict[str, Any]:
        """Generates a JSON schema for a model class.

        Args:
            by_alias: Whether to use attribute aliases or not.
            ref_template: The reference template.
            schema_generator: Overriding the logic used to generate the JSON schema
            mode: The mode in which to generate the schema.
            union_format: The format to use when combining schemas from unions together.


        Returns:
            The JSON schema for the given model class.
        """
        return super().model_json_schema(
            by_alias=by_alias,
            ref_template=ref_template,
            schema_generator=schema_generator,
            mode=mode,
            union_format=union_format,
        )

    @classmethod
    def _convert_non_list_to_list(cls, field_name: str, value: Any) -> list[Any] | None:  # noqa: ANN401
        """Convert a non-list value to a list value by wrapping it in a list."""
        if value is None:
            if field_name in get_field_names_allowing_none(cls):
                return None
            # if a list is required, we interpret None as an empty list
            return []
        # if the value is non-None, wrap it in a list
        return [value]

    @classmethod
    def _convert_list_to_non_list(cls, field_name: str, value: list[Any]) -> Any:  # noqa: ANN401
        """Convert a list value to a non-list value by unpacking it if possible."""
        length = len(value)
        if length == 0:
            # the field might still be required, but that is validated later
            return None
        if length == 1:
            # if we have just one entry, we can safely unpack it
            return value[0]
        # we cannot unambiguously unpack more than one value
        msg = f"got multiple values for {field_name}"
        raise ValueError(msg)

    @classmethod
    def _fix_value_listyness_for_field(cls, field_name: str, value: Any) -> Any:  # noqa: ANN401
        """Check actual and desired shape of a value and fix it if necessary."""
        should_be_list = field_name in get_list_field_names(cls)
        is_list = isinstance(value, list)
        if not is_list and should_be_list:
            return cls._convert_non_list_to_list(field_name, value)
        if is_list and not should_be_list:
            return cls._convert_list_to_non_list(field_name, value)
        # already desired shape
        return value

    @model_validator(mode="wrap")
    @classmethod
    def verify_computed_field_consistency(
        cls,
        data: Any,  # noqa: ANN401
        handler: ValidatorFunctionWrapHandler,
    ) -> Any:  # noqa: ANN401
        """Validate that parsed values for computed fields are consistent.

        Parsing a dictionary with a value for a computed field that is consistent with
        what that field would have computed anyway is allowed. Omitting values for
        computed fields is perfectly valid as well. However, if the parsed value is
        different from the computed value, a validation error is raised.

        Args:
            data: Raw data or instance to be parsed
            handler: Validator function wrap handler

        Returns:
            data with consistent computed fields.
        """
        if not cls.model_computed_fields:
            # no computed fields: exit early
            return handler(data)
        if isinstance(data, cls):
            # data is a model instance: we can assume no computed field was set,
            # because pydantic would throw an AttributeError if you tried
            return handler(data)
        if not isinstance(data, MutableMapping):
            # data is not a dictionary: we can't "pop" values from that,
            # so we can't safely do a before/after comparison
            msg = (
                "Input should be a valid dictionary, validating other types is not "
                "supported for models with computed fields."
            )
            raise AssertionError(msg)  # noqa: TRY004
        custom_values = {
            field: value
            for field in cls.model_computed_fields
            if (value := data.pop(field, None))
        }
        result = handler(data)
        computed_values = result.model_dump(include=set(custom_values))
        if computed_values != custom_values:
            msg = "Cannot set computed fields to custom values!"
            raise ValueError(msg)
        return result

    @model_validator(mode="wrap")
    @classmethod
    def fix_listyness(cls, data: Any, handler: ValidatorFunctionWrapHandler) -> Any:  # noqa: ANN401
        """Adjust the listyness of to-be-parsed data to match the desired shape.

        If that data is a Mapping and the model defines a list[T] field but the raw data
        contains just a value of type T, it will be wrapped into a list. If the raw
        data contains a literal `None`, but the list field is defined as required, we
        substitute an empty list.

        If the model does not expect a list, but the raw data contains a list with
        no entries, it will be substituted with `None`. If the raw data contains exactly
        one entry, then it will be unpacked from the list. If it contains more than one
        entry however, an error is raised, because we would not know which to choose.

        Args:
            data: Raw data or instance to be parsed
            handler: Validator function wrap handler

        Returns:
            data with fixed list shapes
        """
        # TODO(ND): This needs to be a "wrap" validator that is defined *after* the
        # computed field model validator, so it runs *before* the computed field
        # validator. Sigh, see https://github.com/pydantic/pydantic/discussions/7434
        if isinstance(data, MutableMapping):
            for name, value in data.items():
                field_name = get_alias_lookup(cls).get(name, name)
                if field_name in get_all_fields(cls):
                    data[name] = cls._fix_value_listyness_for_field(field_name, value)
        return handler(data)

    def __str__(self) -> str:
        """Format this model as a string with its hash for identification."""
        return f"{self.__class__.__name__}(hash='{hex(hash(self))}')"

    def __hash__(self) -> int:
        """Calculate a hash to make the model usable in sets, dicts and caches.

        Creates a blake2b hash of the model by serializing it with pickle
        and computing a hash digest.

        Returns:
            Hash representing the model's current state.
        """
        serialized = pickle.dumps(self, protocol=pickle.HIGHEST_PROTOCOL)
        digest = hashlib.blake2b(serialized, usedforsecurity=False)
        return int(digest.hexdigest(), 16)



from os import PathLike
from pathlib import Path
from typing import Any, Union

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema


class PathWrapper(PathLike[str]):
    """Custom path for settings that can be absolute or relative to another setting."""

    _path: Path

    def __init__(self, path: Union[str, Path, "PathWrapper"]) -> None:
        """Create a new resolved path instance."""
        if isinstance(path, str):
            path = Path(path)
        elif isinstance(path, PathWrapper):
            path = path._path  # noqa: SLF001
        self._path = path

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """Modify the core schema to add validation and serialization rules."""
        return core_schema.chain_schema(
            [
                core_schema.is_instance_schema(str | Path | PathWrapper),
                core_schema.no_info_plain_validator_function(cls),
            ],
            serialization=core_schema.to_string_ser_schema(when_used="unless-none"),
        )

    def __fspath__(self) -> str:
        """Return the file system path representation."""
        return self._path.__fspath__()

    def __truediv__(self, other: str | PathLike[str]) -> Path:
        """Return a joined path on the basis of `/`."""
        return self._path.__truediv__(other)

    def __hash__(self) -> int:
        """Return the hash for this object."""
        return hash(self._path)

    def __str__(self) -> str:
        """Return a string rendering of the resolved path."""
        return self._path.as_posix()

    def __repr__(self) -> str:
        """Return a representation string of the resolved path."""
        return f'{self.__class__.__name__}("{self}")'

    def __eq__(self, other: object) -> bool:
        """Return true for two PathWrappers with equal paths."""
        if isinstance(other, PathWrapper):
            return self._path.__eq__(other._path)
        msg = f"Can't compare {type(other)} with {type(self)}"
        raise TypeError(msg)

    def is_absolute(self) -> bool:
        """True if the underlying path is absolute."""
        return self._path.is_absolute()

    def is_relative(self) -> bool:
        """True if the underlying path is relative."""
        return not self._path.is_absolute()


class AssetsPath(PathWrapper):
    """Custom path for settings that can be absolute or relative to `assets_dir`."""


class WorkPath(PathWrapper):
    """Custom path for settings that can be absolute or relative to `work_dir`."""



from functools import lru_cache

from mex.common.backend_api.connector import BackendApiConnector
from mex.common.identity.base import BaseProvider
from mex.common.identity.models import Identity
from mex.common.types import Identifier, MergedPrimarySourceIdentifier

IDENTITY_CACHE_SIZE = 5000


class BackendApiIdentityProvider(BaseProvider):
    """Identity provider that communicates with the backend HTTP API."""

    def __init__(self) -> None:
        """Create a new backend identity provider."""
        # mitigating https://docs.astral.sh/ruff/rules/cached-instance-method
        self._cached_assign = lru_cache(IDENTITY_CACHE_SIZE)(self._do_assign)

    def assign(
        self,
        had_primary_source: MergedPrimarySourceIdentifier,
        identifier_in_primary_source: str,
    ) -> Identity:
        """Return a cached Identity from the backend."""
        return self._cached_assign(had_primary_source, identifier_in_primary_source)

    def _do_assign(
        self,
        had_primary_source: MergedPrimarySourceIdentifier,
        identifier_in_primary_source: str,
    ) -> Identity:
        """Find an Identity in the backend or let the backend assign a new one."""
        connector = BackendApiConnector.get()
        return connector.assign_identity(
            had_primary_source=had_primary_source,
            identifier_in_primary_source=identifier_in_primary_source,
        )

    def fetch(
        self,
        *,
        had_primary_source: Identifier | None = None,
        identifier_in_primary_source: str | None = None,
        stable_target_id: Identifier | None = None,
    ) -> list[Identity]:
        """Find Identity instances matching the given filters.

        Either provide `stableTargetId` or `hadPrimarySource`
        and `identifierInPrimarySource` together to get a unique result.
        """
        connector = BackendApiConnector.get()
        response = connector.fetch_identities(
            had_primary_source=had_primary_source,
            identifier_in_primary_source=identifier_in_primary_source,
            stable_target_id=stable_target_id,
        )
        return response.items

    def metrics(self) -> dict[str, int]:
        """Generate metrics about identity provider usage."""
        cache_info = self._cached_assign.cache_info()
        return {"cache_hits": cache_info.hits, "cache_misses": cache_info.misses}

    def close(self) -> None:
        """Clear the connector cache."""
        self._cached_assign.cache_clear()



from typing import Final, Literal, get_args

from mex.common.types.identifier import (
    IDENTIFIER_PATTERN,
    ExtractedAccessPlatformIdentifier,
    ExtractedActivityIdentifier,
    ExtractedBibliographicResourceIdentifier,
    ExtractedConsentIdentifier,
    ExtractedContactPointIdentifier,
    ExtractedDistributionIdentifier,
    ExtractedIdentifier,
    ExtractedOrganizationalUnitIdentifier,
    ExtractedOrganizationIdentifier,
    ExtractedPersonIdentifier,
    ExtractedPrimarySourceIdentifier,
    ExtractedResourceIdentifier,
    ExtractedVariableGroupIdentifier,
    ExtractedVariableIdentifier,
    Identifier,
    MergedAccessPlatformIdentifier,
    MergedActivityIdentifier,
    MergedBibliographicResourceIdentifier,
    MergedConsentIdentifier,
    MergedContactPointIdentifier,
    MergedDistributionIdentifier,
    MergedIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    MergedResourceIdentifier,
    MergedVariableGroupIdentifier,
    MergedVariableIdentifier,
)
from mex.common.types.identity import IdentityProvider
from mex.common.types.link import Link, LinkLanguage
from mex.common.types.path import AssetsPath, PathWrapper, WorkPath
from mex.common.types.sink import Sink
from mex.common.types.temporal_entity import (
    CET,
    TEMPORAL_ENTITY_CLASSES_BY_PRECISION,
    TEMPORAL_ENTITY_FORMATS_BY_PRECISION,
    UTC,
    TemporalEntity,
    TemporalEntityPrecision,
    Year,
    YearMonth,
    YearMonthDay,
    YearMonthDayTime,
)
from mex.common.types.text import Text, TextLanguage
from mex.common.types.validation import Validation
from mex.common.types.vocabulary import (
    VOCABULARY_PATTERN,
    AccessRestriction,
    ActivityType,
    AnonymizationPseudonymization,
    APIType,
    BibliographicResourceType,
    ConsentStatus,
    ConsentType,
    DataProcessingState,
    Frequency,
    Language,
    License,
    MIMEType,
    PersonalData,
    ResourceCreationMethod,
    ResourceTypeGeneral,
    TechnicalAccessibility,
    Theme,
    VocabularyEnum,
    VocabularyLoader,
)

__all__ = (
    "CET",
    "EXTRACTED_IDENTIFIER_CLASSES",
    "EXTRACTED_IDENTIFIER_CLASSES_BY_NAME",
    "IDENTIFIER_PATTERN",
    "MERGED_IDENTIFIER_CLASSES",
    "MERGED_IDENTIFIER_CLASSES_BY_NAME",
    "NESTED_MODEL_CLASSES",
    "NESTED_MODEL_CLASSES_BY_NAME",
    "TEMPORAL_ENTITY_CLASSES_BY_PRECISION",
    "TEMPORAL_ENTITY_FORMATS_BY_PRECISION",
    "UTC",
    "VOCABULARY_ENUMS",
    "VOCABULARY_ENUMS_BY_NAME",
    "VOCABULARY_PATTERN",
    "APIType",
    "AccessRestriction",
    "ActivityType",
    "AnonymizationPseudonymization",
    "AnyExtractedIdentifier",
    "AnyMergedIdentifier",
    "AnyNestedModel",
    "AnyPrimitiveType",
    "AnyValidation",
    "AnyVocabularyEnum",
    "AssetsPath",
    "BibliographicResourceType",
    "ConsentStatus",
    "ConsentType",
    "DataProcessingState",
    "ExtractedAccessPlatformIdentifier",
    "ExtractedActivityIdentifier",
    "ExtractedBibliographicResourceIdentifier",
    "ExtractedConsentIdentifier",
    "ExtractedContactPointIdentifier",
    "ExtractedDistributionIdentifier",
    "ExtractedIdentifier",
    "ExtractedOrganizationIdentifier",
    "ExtractedOrganizationalUnitIdentifier",
    "ExtractedPersonIdentifier",
    "ExtractedPrimarySourceIdentifier",
    "ExtractedResourceIdentifier",
    "ExtractedVariableGroupIdentifier",
    "ExtractedVariableIdentifier",
    "Frequency",
    "Identifier",
    "IdentityProvider",
    "Language",
    "License",
    "Link",
    "LinkLanguage",
    "LiteralStringType",
    "MIMEType",
    "MergedAccessPlatformIdentifier",
    "MergedActivityIdentifier",
    "MergedBibliographicResourceIdentifier",
    "MergedConsentIdentifier",
    "MergedContactPointIdentifier",
    "MergedDistributionIdentifier",
    "MergedIdentifier",
    "MergedOrganizationIdentifier",
    "MergedOrganizationalUnitIdentifier",
    "MergedPersonIdentifier",
    "MergedPrimarySourceIdentifier",
    "MergedResourceIdentifier",
    "MergedVariableGroupIdentifier",
    "MergedVariableIdentifier",
    "PathWrapper",
    "PersonalData",
    "ResourceCreationMethod",
    "ResourceTypeGeneral",
    "Sink",
    "TechnicalAccessibility",
    "TemporalEntity",
    "TemporalEntityPrecision",
    "Text",
    "TextLanguage",
    "Theme",
    "Validation",
    "VocabularyEnum",
    "VocabularyLoader",
    "WorkPath",
    "Year",
    "YearMonth",
    "YearMonthDay",
    "YearMonthDayTime",
)

AnyVocabularyEnum = (
    AccessRestriction
    | ActivityType
    | AnonymizationPseudonymization
    | APIType
    | BibliographicResourceType
    | ConsentStatus
    | ConsentType
    | DataProcessingState
    | Frequency
    | Language
    | License
    | MIMEType
    | PersonalData
    | ResourceCreationMethod
    | ResourceTypeGeneral
    | TechnicalAccessibility
    | Theme
)
VOCABULARY_ENUMS: Final[list[type[AnyVocabularyEnum]]] = list(
    get_args(AnyVocabularyEnum)
)
VOCABULARY_ENUMS_BY_NAME: Final[dict[str, type[AnyVocabularyEnum]]] = {
    cls.__name__: cls for cls in VOCABULARY_ENUMS
}

AnyTemporalEntity = Year | YearMonth | YearMonthDay | YearMonthDayTime
TEMPORAL_ENTITIES: Final[list[type[AnyTemporalEntity]]] = list(
    get_args(AnyTemporalEntity)
)
TEMPORAL_ENTITIES_BY_NAME: Final[dict[str, type[AnyTemporalEntity]]] = {
    cls.__name__: cls for cls in TEMPORAL_ENTITIES
}

AnyNestedModel = Link | Text
NESTED_MODEL_CLASSES: Final[list[type[AnyNestedModel]]] = list(get_args(AnyNestedModel))
NESTED_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyNestedModel]]] = {
    cls.__name__: cls for cls in NESTED_MODEL_CLASSES
}

AnyMergedIdentifier = (
    MergedAccessPlatformIdentifier
    | MergedActivityIdentifier
    | MergedBibliographicResourceIdentifier
    | MergedConsentIdentifier
    | MergedContactPointIdentifier
    | MergedDistributionIdentifier
    | MergedOrganizationalUnitIdentifier
    | MergedOrganizationIdentifier
    | MergedPersonIdentifier
    | MergedPrimarySourceIdentifier
    | MergedResourceIdentifier
    | MergedVariableGroupIdentifier
    | MergedVariableIdentifier
)
MERGED_IDENTIFIER_CLASSES: Final[list[type[AnyMergedIdentifier]]] = list(
    get_args(AnyMergedIdentifier)
)
MERGED_IDENTIFIER_CLASSES_BY_NAME: Final[dict[str, type[AnyMergedIdentifier]]] = {
    cls.__name__: cls for cls in MERGED_IDENTIFIER_CLASSES
}

AnyExtractedIdentifier = (
    ExtractedAccessPlatformIdentifier
    | ExtractedActivityIdentifier
    | ExtractedBibliographicResourceIdentifier
    | ExtractedConsentIdentifier
    | ExtractedContactPointIdentifier
    | ExtractedDistributionIdentifier
    | ExtractedOrganizationalUnitIdentifier
    | ExtractedOrganizationIdentifier
    | ExtractedPersonIdentifier
    | ExtractedPrimarySourceIdentifier
    | ExtractedResourceIdentifier
    | ExtractedVariableGroupIdentifier
    | ExtractedVariableIdentifier
)
EXTRACTED_IDENTIFIER_CLASSES: Final[list[type[AnyExtractedIdentifier]]] = list(
    get_args(AnyExtractedIdentifier)
)
EXTRACTED_IDENTIFIER_CLASSES_BY_NAME: Final[dict[str, type[AnyExtractedIdentifier]]] = {
    cls.__name__: cls for cls in EXTRACTED_IDENTIFIER_CLASSES
}

AnyValidation = Literal[Validation.STRICT, Validation.LENIENT, Validation.IGNORE]
AnyPrimitiveType = str | int | float | None | bool
LiteralStringType = type(Literal["str"])



from typing import Annotated

from pydantic import BaseModel, Field


class FilterRule(BaseModel, extra="forbid"):
    """A single filter rule to apply."""

    forValues: Annotated[list[str] | None, Field(title="forValues")] = None
    rule: Annotated[str | None, Field(title="rule")] = None


class FilterField(BaseModel, extra="forbid"):
    """Filter definition for one field in the primary source."""

    fieldInPrimarySource: Annotated[str | None, Field(title="fieldInPrimarySource")] = (
        None
    )
    locationInPrimarySource: Annotated[
        str | None, Field(title="locationInPrimarySource")
    ] = None
    examplesInPrimarySource: Annotated[
        list[str] | None, Field(title="examplesInPrimarySource")
    ] = None
    filterRules: Annotated[list[FilterRule], Field(min_length=1, title="filterRules")]
    comment: Annotated[str | None, Field(title="comment")] = None


class BaseFilter(BaseModel, extra="forbid"):
    """Base class for filter implementations."""



import json
from collections.abc import Generator, Iterable
from contextlib import ExitStack
from pathlib import Path
from typing import IO, Any, TypeVar

from mex.common.logging import logger, watch
from mex.common.models import BaseModel
from mex.common.settings import BaseSettings
from mex.common.sinks.base import BaseSink
from mex.common.transform import MExEncoder
from mex.common.utils import grouper

_LoadItemT = TypeVar("_LoadItemT", bound=BaseModel)


class NdjsonSink(BaseSink):
    """Sink to load models into new-line delimited JSON files."""

    CHUNK_SIZE = 100
    _work_dir: Path

    def __init__(self) -> None:
        """Instantiate the multi sink singleton."""
        settings = BaseSettings.get()
        self._work_dir = Path(settings.work_dir)

    def close(self) -> None:
        """Nothing to close, since load already closes all file handles."""

    @watch(log_interval=1000)
    def load(
        self,
        items: Iterable[_LoadItemT],
    ) -> Generator[_LoadItemT, None, None]:
        """Write any items into a new-line delimited JSON file.

        Args:
            items: Iterable of any kind of items

        Returns:
            Generator for the loaded items
        """
        file_handles: dict[str, IO[Any]] = {}
        with ExitStack() as stack:
            for chunk in grouper(self.CHUNK_SIZE, items):
                for item in chunk:
                    if item is None:
                        continue
                    class_name = item.__class__.__name__
                    try:
                        fh = file_handles[class_name]
                    except KeyError:
                        file = self._work_dir / f"{class_name}.ndjson"
                        writer = file.open("a+", encoding="utf-8")
                        file_handles[class_name] = fh = stack.enter_context(writer)
                        logger.info(
                            "%s - writing %s to file %s",
                            type(self).__name__,
                            class_name,
                            file.as_posix(),
                        )
                    dumped_json = json.dumps(
                        item, sort_keys=True, cls=MExEncoder, ensure_ascii=False
                    )
                    fh.write(f"{dumped_json}\n")
                    yield item



from collections.abc import Iterable

from mex.common.logging import logger
from mex.common.models import (
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    ExtractedPrimarySource,
)
from mex.common.primary_source.models import SeedPrimarySource


def transform_seed_primary_source_to_extracted_primary_source(
    primary_source: SeedPrimarySource,
) -> ExtractedPrimarySource:
    """Transform a seed primary source into an ExtractedPrimarySource.

    Args:
        primary_source: Primary source coming from raw-data file

    Returns:
        ExtractedPrimarySource
    """
    return ExtractedPrimarySource(
        identifierInPrimarySource=primary_source.identifier,
        title=primary_source.title,
        hadPrimarySource=MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    )


def transform_seed_primary_sources_to_extracted_primary_sources(
    seed_primary_sources: Iterable[SeedPrimarySource],
) -> list[ExtractedPrimarySource]:
    """Transform seed primary sources into ExtractedPrimarySources.

    Args:
        seed_primary_sources: Iterable of primary sources coming from raw-data file

    Returns:
        List of ExtractedPrimarySource
    """
    extracted_primary_sources = [
        transform_seed_primary_source_to_extracted_primary_source(primary_source)
        for primary_source in seed_primary_sources
    ]
    logger.info("transformed %s primary sources", len(extracted_primary_sources))
    return extracted_primary_sources



import re
from functools import lru_cache

from mex.common.wikidata.connector import WikidataAPIConnector
from mex.common.wikidata.models import WikidataOrganization


@lru_cache(maxsize=128)
def get_wikidata_organization(item_id_or_url: str) -> WikidataOrganization:
    """Get a wikidata item details by its ID.

    Args:
        item_id_or_url: Wikidata item ID or full URL

    Raises:
        ValueError: when item_id_or_url does not match pattern

    Returns:
        WikidataOrganization object
    """
    if match := re.search(
        r"^(?:http?://)?(?:www\.)?(?:wikidata\.org/entity/)?([A-Z0-9]{2,64})/?$",
        item_id_or_url,
    ):
        item_id = match.group(1)
    else:
        msg = f"malformed wikidata url: {item_id_or_url}"
        raise ValueError(msg)
    connector = WikidataAPIConnector.get()
    item = connector.get_wikidata_item_details_by_id(item_id)
    return WikidataOrganization.model_validate(item)



from enum import StrEnum
from typing import Annotated, Any

from langdetect.detector_factory import PROFILES_DIRECTORY, DetectorFactory
from langdetect.lang_detect_exception import LangDetectException
from pydantic import BaseModel, Field, model_validator

DETECTOR_FACTORY = DetectorFactory()
DETECTOR_FACTORY.load_profile(PROFILES_DIRECTORY)
DETECTOR_FACTORY.seed = 0


class TextLanguage(StrEnum):
    """Possible language tags for `Text` values."""

    DE = "de"
    EN = "en"
    FR = "fr"
    ES = "es"
    RU = "ru"


class RestrictedTextLanguage(StrEnum):
    """Allows only English and German as language tags for `Text` values."""

    DE = "de"
    EN = "en"


def get_language_by_confidence(
    detector: DetectorFactory, confidence_threshold: float = 0.75
) -> TextLanguage | None:
    """Assigns None as Language if confidence is below 0.75 or is not En or DE."""
    probs = detector.get_probabilities()
    if not probs:
        return None

    best = max(probs, key=lambda p: p.prob)

    if best.prob < confidence_threshold:
        return None
    try:
        return TextLanguage(RestrictedTextLanguage(best.lang).value)
    except ValueError:
        return None


class Text(BaseModel):
    """Type class for text objects.

    Texts can be parsed from nested JSON objects or from raw strings.

    Example:
        Text(value="foo") == Text.model_validate("foo")
    """

    value: Annotated[str, Field(min_length=1)]
    language: TextLanguage | None = None

    @model_validator(mode="before")
    @classmethod
    def detect_language(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Detect the language of the text if not explicitly given."""
        language = values.get("language")
        value = values.get("value")
        if value and "language" not in values:
            try:
                detector = DETECTOR_FACTORY.create()
                detector.append(value)
                language = get_language_by_confidence(detector)
            except (LangDetectException, ValueError):
                pass
        return {"language": language, "value": value}

    @model_validator(mode="before")
    @classmethod
    def validate_strings(cls, value: Any) -> dict[str, Any]:  # noqa: ANN401
        """Convert string input to dictionary."""
        if isinstance(value, str):
            return {"value": value}
        if isinstance(value, dict):
            return value
        msg = f"Allowed input types are dict and str, got {type(value)}"
        raise ValueError(msg)

    def __hash__(self) -> int:
        """Return the hash of Text."""
        return hash((self.value, self.language))



import json
import time
from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Literal, cast

import backoff
import requests
from requests import RequestException, Response, codes
from requests.exceptions import (
    ConnectTimeout,
    ProxyError,
    ReadTimeout,
    SSLError,
)

from mex.common.connector import BaseConnector
from mex.common.connector.utils import bounded_backoff
from mex.common.exceptions import (
    TimedReadTimeout,
    TimedServerError,
    TimedTooManyRequests,
)
from mex.common.logging import logger
from mex.common.settings import BaseSettings
from mex.common.transform import MExEncoder


class HTTPConnector(BaseConnector):
    """Base class for requests-based HTTP connectors."""

    TIMEOUT: int | float = 10
    TIMEOUT_MAX: int | float = 100
    PROPORTIONAL_BACKOFF_MIN: int | float = 3
    url: str = ""

    def __init__(self) -> None:
        """Create a new http connection."""
        self._set_session()
        self._set_url()
        self._set_authentication()
        self._check_availability()

    def _set_session(self) -> None:
        """Create and set request session."""
        settings = BaseSettings.get()
        self.session = requests.Session()
        self.session.verify = settings.verify_session  # type: ignore[assignment]

    def _set_authentication(self) -> None:
        """Authenticate to the host."""

    @abstractmethod
    def _set_url(self) -> None:
        """Set url of the host."""

    def _check_availability(self) -> None:
        """Send a GET request to verify the host is available."""
        self._send_request("HEAD", self.url, params={})

    @backoff.on_exception(  # try to overcome network issues
        wait_gen=backoff.fibo,
        exception=(ConnectTimeout, ProxyError, SSLError),
        max_tries=3,
        jitter=backoff.random_jitter,
        logger=logger,
    )
    @backoff.on_exception(  # proportionally backoff on server fault
        wait_gen=backoff.runtime,
        exception=(TimedReadTimeout, TimedTooManyRequests, TimedServerError),
        value=bounded_backoff(PROPORTIONAL_BACKOFF_MIN, TIMEOUT_MAX),
        max_tries=5,
        jitter=backoff.random_jitter,
        logger=logger,
    )
    def _send_request(
        self,
        method: str,
        url: str,
        params: Mapping[str, list[str] | str | None] | None,
        **kwargs: Any,  # noqa: ANN401
    ) -> Response:
        """Send the request with advanced retrying rules."""
        t0 = time.perf_counter()
        try:
            response = self.session.request(method, url, params, **kwargs)
            response.raise_for_status()
        except ReadTimeout as exc:
            raise TimedReadTimeout.create(exc, t0) from exc
        except RequestException as exc:
            if exc.response and exc.response.status_code >= codes.internal_server_error:
                raise TimedServerError.create(exc, t0) from exc
            if exc.response and exc.response.status_code == codes.too_many_requests:
                raise TimedTooManyRequests.create(exc, t0) from exc
            raise
        return response

    def request_raw(
        self,
        method: Literal["OPTIONS", "POST", "GET", "PUT", "DELETE"],
        endpoint: str | None = None,
        payload: Any = None,  # noqa: ANN401
        params: Mapping[str, list[str] | str | None] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> Response:
        """Prepare and send a raw request with error handling and payload serialization.

        Args:
            method: HTTP method to use
            endpoint: Path to API endpoint to be prefixed with host and version
            payload: Data to be serialized as JSON using the `MExEncoder`
            params: Dictionary to be sent in the query string of the request
            kwargs: Further keyword arguments passed to `requests`

        Raises:
            RequestException: Error from `requests` that can't be solved with a retry
            HTTPError: Re-raised HTTP error with (truncated) response body

        Returns:
            Response object for the  request
        """
        # Prepare request
        if endpoint:
            url = f"{self.url.rstrip('/')}/{endpoint.lstrip('/')}"
        else:
            url = self.url
        kwargs.setdefault("timeout", self.TIMEOUT)
        if not kwargs.get("headers"):
            kwargs.setdefault("headers", {})
        kwargs["headers"].setdefault("User-Agent", "rki/mex")

        if payload:
            kwargs["data"] = json.dumps(payload, cls=MExEncoder)

        # Send request
        return self._send_request(method, url, params, **kwargs)

    def request(
        self,
        method: Literal["OPTIONS", "POST", "GET", "PUT", "DELETE"],
        endpoint: str | None = None,
        payload: Any = None,  # noqa: ANN401
        params: Mapping[str, list[str] | str | None] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> dict[str, Any]:
        """Prepare and send a request with error handling and payload de/serialization.

        Args:
            method: HTTP method to use
            endpoint: Path to API endpoint to be prefixed with host and version
            payload: Data to be serialized as JSON using the `MExEncoder`
            params: Dictionary to be sent in the query string of the request
            kwargs: Further keyword arguments passed to `requests`

        Raises:
            RequestException: Error from `requests` that can't be solved with a retry
            HTTPError: Re-raised HTTP error with (truncated) response body
            JSONDecodeError: If body of response cannot be parsed correctly

        Returns:
            Parsed JSON body of the response
        """
        if not kwargs.get("headers"):
            kwargs.setdefault("headers", {})
        kwargs["headers"].setdefault("Accept", "application/json")
        response = self.request_raw(
            method, endpoint=endpoint, payload=payload, params=params, **kwargs
        )
        if response.status_code == codes.no_content:
            return {}
        return cast("dict[str, Any]", response.json())

    def close(self) -> None:
        """Close the connector's underlying requests session."""
        self.session.close()



from collections.abc import Iterator
from typing import Generic, TypeVar

_SingletonT = TypeVar("_SingletonT")


class SingletonStore(Generic[_SingletonT]):
    """Thin wrapper for storing thread-local singletons."""

    def __init__(self) -> None:
        """Create a new singleton store with the given type."""
        self._instances_by_class: dict[type[_SingletonT], _SingletonT] = {}

    def __iter__(self) -> Iterator[_SingletonT]:
        """Return an iterator over all singletons in the store."""
        return iter(self._instances_by_class.values())

    def load(self, cls: type[_SingletonT]) -> _SingletonT:
        """Retrieve a singleton for the given class or create a new one."""
        if cls not in self._instances_by_class:
            self.push(cls())
        return self._instances_by_class[cls]

    def push(self, instance: _SingletonT) -> None:
        """Set or replace a singleton instance in the store."""
        self._instances_by_class[type(instance)] = instance

    def pop(self, cls: type[_SingletonT]) -> _SingletonT:
        """Remove a singleton for the given class for the store and return it."""
        return self._instances_by_class.pop(cls)

    def reset(self) -> None:
        """Remove all singleton instances from the store."""
        self._instances_by_class.clear()


class SingleSingletonStore(Generic[_SingletonT]):
    """Thin wrapper for storing a single thread-local singleton.

    Stores only a single instance. Requested class must either be
    the same or a parent class of the stored class.
    """

    def __init__(self) -> None:
        """Create a new settings singleton store."""
        self._singleton: _SingletonT | None = None

    def load(self, cls: type[_SingletonT]) -> _SingletonT:
        """Retrieve the settings for the given class or create a new one."""
        if self._singleton is None:
            self._singleton = cls()
            return self._singleton
        if not issubclass(type(self._singleton), cls):
            msg = (
                f"requested class ({cls}) is not a parent class of loaded class "
                f"({type(self._singleton)}). "
                f"Did you initialize {cls} upon startup?"
            )
            raise RuntimeError(msg)  # noqa: TRY004
        return self._singleton

    def push(self, instance: _SingletonT) -> None:
        """Set or replace a singleton instance in the store."""
        self._singleton = instance

    def reset(self) -> None:
        """Remove singleton instance from the store."""
        self._singleton = None



from collections.abc import Generator, Iterable, Sequence

from mex.common.models import ExtractedOrganization
from mex.common.types import MergedPrimarySourceIdentifier, Text, TextLanguage
from mex.common.wikidata.models import Aliases, Claim, Labels, WikidataOrganization


def transform_wikidata_organizations_to_extracted_organizations(
    wikidata_organizations: Iterable[WikidataOrganization],
    wikidata_primary_source_id: MergedPrimarySourceIdentifier,
) -> Generator[ExtractedOrganization, None, None]:
    """Transform wikidata organizations into ExtractedOrganizations.

    Wikidata organizations without labels are skipped.

    Args:
        wikidata_organizations: Iterable of wikidata organization to be transformed
        wikidata_primary_source_id: Extracted primary source id for wikidata

    Returns:
        Generator of ExtractedOrganizations
    """
    for wikidata_organization in wikidata_organizations:
        if extracted_organization := (
            transform_wikidata_organization_to_extracted_organization(
                wikidata_organization, wikidata_primary_source_id
            )
        ):
            yield extracted_organization


def transform_wikidata_organization_to_extracted_organization(
    wikidata_organization: WikidataOrganization,
    wikidata_primary_source_id: MergedPrimarySourceIdentifier,
) -> ExtractedOrganization | None:
    """Transform one wikidata organization into ExtractedOrganizations.

    If no labels are found on the wikidata organization, `None` is returned instead.

    Args:
        wikidata_organization: wikidata organization to be transformed
        wikidata_primary_source_id: Extracted primary source id for wikidata

    Returns:
        ExtractedOrganization or None
    """
    labels = get_official_name_label(wikidata_organization.labels)
    if not labels:
        return None
    return ExtractedOrganization(
        wikidataId=f"http://www.wikidata.org/entity/{wikidata_organization.identifier}",
        officialName=labels,
        shortName=_get_clean_short_names(wikidata_organization.claims.short_name),
        geprisId=[],
        isniId=[
            f"https://isni.org/isni/{claim.mainsnak.datavalue.value.text}".replace(
                " ", ""
            )
            for claim in wikidata_organization.claims.isni_id
        ],
        gndId=[
            f"https://d-nb.info/gnd/{claim.mainsnak.datavalue.value.text}"
            for claim in wikidata_organization.claims.gnd_id
        ],
        viafId=[
            f"https://viaf.org/viaf/{claim.mainsnak.datavalue.value.text}"
            for claim in wikidata_organization.claims.viaf_id
        ],
        rorId=[
            f"https://ror.org/{claim.mainsnak.datavalue.value.text}"
            for claim in wikidata_organization.claims.ror_id
        ],
        identifierInPrimarySource=wikidata_organization.identifier,
        hadPrimarySource=wikidata_primary_source_id,
        alternativeName=_get_alternative_names(
            wikidata_organization.claims.native_label, wikidata_organization.aliases
        ),
    )


def _get_alternative_names(
    native_labels: Sequence[Claim],
    all_aliases: Aliases,
) -> list[Text]:
    """Get alternative names of an organization in DE and EN.

    Args:
        native_labels: Sequence of all native labels
        all_aliases: All aliases of the organization

    Returns:
        combined list of native labels and aliases in DE and EN
    """
    alternative_names = []

    for alias in all_aliases.en + all_aliases.de:
        text = Text(value=alias.value, language=None)
        if text not in alternative_names:
            alternative_names.append(text)

    for native_label in native_labels:
        value = native_label.mainsnak.datavalue.value.text
        language = native_label.mainsnak.datavalue.value.language

        if not value:
            continue

        if language == "de":
            text = Text(value=value, language=TextLanguage.DE)
        elif language == "en":
            text = Text(value=value, language=TextLanguage.EN)
        else:
            continue

        if text not in alternative_names:
            alternative_names.append(text)

    return alternative_names


def _get_clean_short_names(short_names: Sequence[Claim]) -> list[Text]:
    """Get clean short names only in EN and DE and ignore the rest.

    Args:
        short_names: List of all short names

    Returns:
        list of clean short names in EN and DE
    """
    clean_short_names = []
    for short_name in short_names:
        value = short_name.mainsnak.datavalue.value.text
        language = short_name.mainsnak.datavalue.value.language

        if not value:
            continue

        if language == "de":
            text = Text(value=value, language=TextLanguage.DE)
        elif language == "en":
            text = Text(value=value, language=TextLanguage.EN)
        else:
            continue

        if text not in clean_short_names:
            clean_short_names.append(text)

    return clean_short_names


def get_official_name_label(labels: Labels) -> Text | None:
    """Get if DE label is available and return a list of EN and DE labels.

    Args:
        labels: Wikidata labels object

    Returns:
        Text object of the label that was picked, or None
    """
    if labels.de:
        return Text(value=labels.de.value, language=TextLanguage.DE)
    if labels.en:
        return Text(value=labels.en.value, language=TextLanguage.EN)
    if labels.multiple:
        return Text(value=labels.multiple.value, language=None)
    return None



import re
from collections.abc import Iterable
from dataclasses import dataclass
from functools import lru_cache

from mex.common.ldap.models import (
    AnyLDAPActor,
    LDAPFunctionalAccount,
    LDAPPerson,
    LDAPPersonWithQuery,
)
from mex.common.logging import logger
from mex.common.models import (
    ExtractedContactPoint,
    ExtractedOrganization,
    ExtractedOrganizationalUnit,
    ExtractedPerson,
)
from mex.common.types import (
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPrimarySourceIdentifier,
)
from mex.common.utils import deprecated


def transform_ldap_persons_to_extracted_persons(
    ldap_persons: Iterable[LDAPPerson],
    primary_source_id: MergedPrimarySourceIdentifier,
    units: Iterable[ExtractedOrganizationalUnit],
    rki_organization: ExtractedOrganization,
) -> list[ExtractedPerson]:
    """Transform LDAP persons to ExtractedPersons.

    Args:
        ldap_persons: LDAP persons
        primary_source_id: Primary source identifier for LDAP
        units: Extracted organizational units
        rki_organization: RKI organization to which persons belong

    Returns:
        List of extracted persons
    """
    units_by_identifier_in_primary_source = {
        unit.identifierInPrimarySource: unit for unit in units
    }
    extracted_persons = [
        transform_ldap_person_to_extracted_person(
            person,
            primary_source_id,
            units_by_identifier_in_primary_source,
            rki_organization,
        )
        for person in ldap_persons
    ]
    logger.info("transformed %s extracted persons from ldap", len(extracted_persons))
    return extracted_persons


def transform_ldap_functional_accounts_to_extracted_contact_points(
    ldap_functional_accounts: Iterable[LDAPFunctionalAccount],
    primary_source_id: MergedPrimarySourceIdentifier,
) -> list[ExtractedContactPoint]:
    """Transform LDAP functional accounts to ExtractedContactPoints.

    Args:
        ldap_functional_accounts: LDAP functional accounts
        primary_source_id: Primary source identifier for LDAP

    Returns:
        List of extracted contact points
    """
    extracted_contact_points = [
        transform_ldap_functional_account_to_extracted_contact_point(
            functional_account, primary_source_id
        )
        for functional_account in ldap_functional_accounts
    ]
    logger.info(
        "transformed %s extracted contact points from ldap",
        len(extracted_contact_points),
    )
    return extracted_contact_points


def transform_ldap_persons_with_query_to_extracted_persons(
    ldap_persons_with_query: Iterable[LDAPPersonWithQuery],
    primary_source_id: MergedPrimarySourceIdentifier,
    units: Iterable[ExtractedOrganizationalUnit],
    rki_organization: ExtractedOrganization,
) -> list[ExtractedPerson]:
    """Transform LDAP persons with query to ExtractedPersons.

    Args:
        ldap_persons_with_query: LDAP persons with query
        primary_source_id: Primary source identifier for LDAP
        units: Extracted organizational units
        rki_organization: RKI organization to which persons belong

    Returns:
        List of extracted persons
    """
    return transform_ldap_persons_to_extracted_persons(
        (a.person for a in ldap_persons_with_query),
        primary_source_id,
        units,
        rki_organization,
    )


def transform_ldap_person_to_extracted_person(
    ldap_person: LDAPPerson,
    primary_source_id: MergedPrimarySourceIdentifier,
    units_by_identifier_in_primary_source: dict[str, ExtractedOrganizationalUnit],
    rki_organization: ExtractedOrganization,
) -> ExtractedPerson:
    """Transform a single LDAP person to an ExtractedPerson.

    Args:
        ldap_person: LDAP person
        primary_source_id: Primary source identifier for LDAP
        units_by_identifier_in_primary_source: Mapping to get units by LDAP ID
        rki_organization: RKI Organization

    Returns:
        Extracted person
    """
    person_unit_ids = [
        unit.stableTargetId
        for d in (ldap_person.department, ldap_person.departmentNumber)
        if d and (unit := units_by_identifier_in_primary_source.get(d.lower()))
    ]
    return transform_ldap_person_and_unit_ids_to_extracted_person(
        ldap_person, primary_source_id, person_unit_ids, rki_organization.stableTargetId
    )


def transform_ldap_person_and_unit_ids_to_extracted_person(
    ldap_person: LDAPPerson,
    primary_source_id: MergedPrimarySourceIdentifier,
    person_unit_ids: list[MergedOrganizationalUnitIdentifier],
    rki_organization_id: MergedOrganizationIdentifier,
) -> ExtractedPerson:
    """Transform a single LDAP person and organizational unit ids to an ExtractedPerson.

    Args:
        ldap_person: LDAP person
        primary_source_id: Primary source identifier for LDAP
        person_unit_ids: ids of units affiliated with person
        rki_organization_id: RKI Organization id

    Returns:
        Extracted person
    """
    return ExtractedPerson(
        identifierInPrimarySource=str(ldap_person.objectGUID),
        hadPrimarySource=primary_source_id,
        affiliation=[rki_organization_id],
        email=ldap_person.mail,
        familyName=[ldap_person.sn],
        fullName=[ldap_person.displayName] if ldap_person.displayName else [],
        givenName=ldap_person.givenName,
        isniId=[],
        memberOf=person_unit_ids,
        orcidId=[],
    )


def transform_ldap_functional_account_to_extracted_contact_point(
    ldap_functional_account: LDAPFunctionalAccount,
    primary_source_id: MergedPrimarySourceIdentifier,
) -> ExtractedContactPoint:
    """Transform a single LDAP functional account to an ExtractedContactPoint.

    Args:
        ldap_functional_account: LDAP functional account
        primary_source_id: Primary source identifier for LDAP

    Returns:
        Extracted contact point
    """
    return ExtractedContactPoint(
        identifierInPrimarySource=str(ldap_functional_account.objectGUID),
        hadPrimarySource=primary_source_id,
        email=ldap_functional_account.mail,
    )


def transform_any_ldap_actor_to_extracted_persons_or_contact_points(
    ldap_actors: Iterable[AnyLDAPActor],
    units: Iterable[ExtractedOrganizationalUnit],
    primary_source_id: MergedPrimarySourceIdentifier,
    rki_organization: ExtractedOrganization,
) -> list[ExtractedPerson | ExtractedContactPoint]:
    """Transform any LDAP actors to extracted person or contact points.

    Args:
        ldap_actors: Iterable of LDAP functional account or LDAP persons
        units: Extracted organizational units
        primary_source_id: Primary source identifier for LDAP
        rki_organization: RKI organization to which persons belong

    Returns:
        Extracted persons or contact points
    """
    units_by_identifier_in_primary_source = {
        unit.identifierInPrimarySource: unit for unit in units
    }
    extracted_actors = [
        transform_ldap_person_to_extracted_person(
            actor,
            primary_source_id,
            units_by_identifier_in_primary_source,
            rki_organization,
        )
        if isinstance(actor, LDAPPerson)
        else transform_ldap_functional_account_to_extracted_contact_point(
            actor, primary_source_id
        )
        for actor in ldap_actors
    ]
    logger.info("transformed %s extracted actors from ldap", len(extracted_actors))
    return extracted_actors


@dataclass
class PersonName:
    """Name of a person split into sur- and given-name."""

    surname: str = "*"
    given_name: str = "*"
    full_name: str = ""


@lru_cache(maxsize=1024)
def analyse_person_string(string: str) -> list[PersonName]:
    """Try to extract a list of given- and surnames from a person string.

    For supported formats of this implementation, check unittest.

    Args:
        string: Person string, containing their name in some form

    Returns:
        List of analyzed person names
    """
    # remove string used to designate no person
    string = string.strip("-")
    # remove everything in brackets
    string = re.sub(r"\([^\)]*[\)$]", "", string)
    # remove tangling brackets
    string = re.sub(r"\(|\)", "", string)
    # remove titles case insensitive
    string = re.sub(r"(hr|fr|dr|med|prof|stellv)\.", "", string, flags=re.IGNORECASE)
    # remove some special keywords
    string = re.sub(r"(Abteilung|Projektleitung|Leitung)", "", string)
    # remove any word with numbers in it
    string = re.sub(r"\b([A-Z][a-zA-Z]*[0-9]+)\b", "", string)
    # remove any word all caps word with dot
    string = re.sub(r"\b([A-Z]+\.)", "", string)
    # remove any remaining numeric characters
    string = re.sub(r"[0-9]", "", string)
    # remove any padding spaces or commas
    string = string.strip(" ,")

    # split on forward slash or semicolon
    if len(split := re.split(r"/|;", string)) > 1:
        return [name for strings in split for name in analyse_person_string(strings)]

    # split on comma if there is more than one
    if len(split := re.split(r",", string)) > 2:  # noqa: PLR2004
        return [name for strings in split for name in analyse_person_string(strings)]

    # split on single commas only if there are more than three words
    if len(split := re.split(r",", string)) == 2 and string.strip().count(" ") > 2:  # noqa: PLR2004
        return [name for strings in split for name in analyse_person_string(strings)]

    # split into surname and given name
    if "," in string:
        # if we have a comma, use that to split
        split = list(reversed(string.split(",", maxsplit=1)))
    else:
        # if not, split on whitespace
        split = string.rsplit(maxsplit=1)

    # normalize multiple or padding whitespaces
    split = [" ".join(s.split()) for s in split if s.strip()]
    # re-assemble full name from split
    full_name = " ".join(split)

    # return only surname when given name is missing
    if len(split) == 1:
        return [PersonName(surname=split[0], full_name=full_name)]

    # return surname and given name
    if len(split) == 2:  # noqa: PLR2004
        return [PersonName(surname=split[1], given_name=split[0], full_name=full_name)]

    # found no one
    return []


# Deprecated old function names
transform_ldap_persons_to_mex_persons = deprecated(
    "transform_ldap_persons_to_mex_persons",
    transform_ldap_persons_to_extracted_persons,
)
transform_ldap_actors_to_mex_contact_points = deprecated(
    "transform_ldap_actors_to_mex_contact_points",
    transform_ldap_functional_accounts_to_extracted_contact_points,
)
transform_ldap_persons_with_query_to_mex_persons = deprecated(
    "transform_ldap_persons_with_query_to_mex_persons",
    transform_ldap_persons_with_query_to_extracted_persons,
)
transform_ldap_person_to_mex_person = deprecated(
    "transform_ldap_person_to_mex_person",
    transform_ldap_person_to_extracted_person,
)
transform_ldap_actor_to_mex_contact_point = deprecated(
    "transform_ldap_actor_to_mex_contact_point",
    transform_ldap_functional_account_to_extracted_contact_point,
)



from mex.common.models.base.model import BaseModel
from mex.common.transform import camel_to_split


class MergedItem(
    BaseModel,
    extra="forbid",
    model_title_generator=lambda m: camel_to_split(m.__name__),
):
    """Base model for all merged item classes."""



from collections.abc import Callable

from mex.common.exceptions import TimedRequestException


def bounded_backoff(
    min_time: float, max_time: float
) -> Callable[[TimedRequestException], int | float]:
    """Get a function to calculate a bounded backoff time.

    Args:
        min_time: Minimum backoff time in seconds.
        max_time: Maximum backoff time in seconds.

    Returns:
        Function that takes a TimedRequestException and returns a bounded delay time.
    """

    def wrapped(error: TimedRequestException) -> int | float:
        return min(max(error.seconds, min_time), max_time)

    return wrapped
